import random
import re
import warnings
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import pandas as pd
from ruamel.yaml import YAML


class Config(object):
    """Read run configuration from the specified path or dictionary and parse it into a configuration object.

    During parsing, config keys that contain 'dir', 'file', or 'path' will be converted to pathlib.Path instances.
    Configuration keys ending with '_date' will be parsed to pd.Timestamps. The expected format is DD/MM/YYYY.

    Parameters
    ----------
    yml_path_or_dict : Union[Path, dict]
        Either a path to the config file or a dictionary of configuration values.
    dev_mode : bool, optional
        If dev_mode is off, the config creation will fail if there are unrecognized keys in the passed config
        specification. dev_mode can be activated either through this parameter or by setting ``dev_mode: True``
        in `yml_path_or_dict`.

    Raises
    ------
    ValueError
        If the passed configuration specification is neither a Path nor a dict or if `dev_mode` is off (default) and
        the config file or dict contain unrecognized keys.
    """

    # Lists of deprecated config keys and purely informational metadata keys, needed when checking for unrecognized
    # config keys since these keys are not properties of the Config class.
    _deprecated_keys = [
        'static_inputs', 'camels_attributes', 'target_variable', 'embedding_hiddens', 'embedding_activation',
        'embedding_dropout'
    ]
    _metadata_keys = ['package_version', 'commit_hash']

    def __init__(self, yml_path_or_dict: Union[Path, dict], dev_mode: bool = False):
        if isinstance(yml_path_or_dict, Path):
            self._cfg = Config._read_and_parse_config(yml_path=yml_path_or_dict)
        elif isinstance(yml_path_or_dict, dict):
            self._cfg = Config._parse_config(yml_path_or_dict)
        else:
            raise ValueError(f'Cannot create a config from input of type {type(yml_path_or_dict)}.')

        if not (self._cfg.get('dev_mode', False) or dev_mode):
            Config._check_cfg_keys(self._cfg)

        # Adjust experiment name
        if "experiment_name" in self._cfg and self._cfg["experiment_name"]:
            # 1. Replace curly-bracketed entries
            new_name = self._cfg["experiment_name"]
            for key, val in self._cfg.items():
                if key.endswith('_date'):
                    if isinstance(val, list):
                        date_string = '_'.join(elem.strftime('%Y-%m-%d') for elem in val)
                    else:
                        date_string = val.strftime('%Y-%m-%d')
                    new_name = re.sub(f'{{{key}}}', date_string, new_name)
            new_name = re.sub('{random_name}', create_random_name(), new_name)
            try:
                new_name = new_name.format(**self._cfg)
            except KeyError as ex:
                raise KeyError(f'Experiment name is {self._cfg["experiment_name"]} ' +
                               f'but {{{ex}}} was not found in config.') from ex
            # 2. Remove curly brackets and make sure experiment name can be used as folder in Linux and Windows
            new_name = re.sub('{', "(", new_name)
            new_name = re.sub('}', ")", new_name)
            new_name = re.sub('"', "'", new_name)
            new_name = re.sub('[<>:"/\\\\|?*]', "_", new_name)
            new_name = re.sub(' ', '', new_name)

            self._cfg["experiment_name"] = new_name

    def as_dict(self) -> dict:
        """Return run configuration as dictionary.
        
        Returns
        -------
        dict
            The run configuration, as defined in the .yml file.
        """
        return self._cfg

    def dump_config(self, folder: Path, filename: str = 'config.yml'):
        """Save the run configuration as a .yml file to disk.

        Parameters
        ----------
        folder : Path
            Folder in which the configuration will be stored.
        filename : str, optional
            Name of the file that will be stored. Default: 'config.yml'.

        Raises
        ------
        FileExistsError
            If the specified folder already contains a file named `filename`.
        """
        yml_path = folder / filename
        if not yml_path.exists():
            with yml_path.open('w') as fp:
                temp_cfg = {}
                for key, val in self._cfg.items():
                    if any([key.endswith(x) for x in ['_dir', '_path', '_file', '_files']]):
                        if isinstance(val, list):
                            temp_list = []
                            for elem in val:
                                temp_list.append(str(elem))
                            temp_cfg[key] = temp_list
                        else:
                            temp_cfg[key] = str(val)
                    elif key.endswith('_date'):
                        if isinstance(val, list):
                            temp_list = []
                            for elem in val:
                                temp_list.append(elem.strftime(format="%d/%m/%Y"))
                            temp_cfg[key] = temp_list
                        else:
                            # Ignore None's due to e.g. using a per_basin_period_file
                            if isinstance(val, pd.Timestamp):
                                temp_cfg[key] = val.strftime(format="%d/%m/%Y")
                    else:
                        temp_cfg[key] = val

                yaml = YAML()
                yaml.dump(dict(OrderedDict(sorted(temp_cfg.items()))), fp)
        else:
            raise FileExistsError(yml_path)

    def update_config(self, yml_path_or_dict: Union[Path, dict], dev_mode: bool = False):
        """Update config arguments.
        
        Useful e.g. in the context of fine-tuning or when continuing to train from a checkpoint to adapt for example the
        learning rate, train basin files or anything else.
        
        Parameters
        ----------
        yml_path_or_dict : Union[Path, dict]
            Either a path to the new config file or a dictionary of configuration values. Each argument specified in
            this file will overwrite the existing config argument.
        dev_mode : bool, optional
            If dev_mode is off, the config creation will fail if there are unrecognized keys in the passed config
            specification. dev_mode can be activated either through this parameter or by setting ``dev_mode: True``
            in `yml_path_or_dict`.

        Raises
        ------
        ValueError
            If the passed configuration specification is neither a Path nor a dict, or if `dev_mode` is off (default)
            and the config file or dict contain unrecognized keys.
        """
        new_config = Config(yml_path_or_dict, dev_mode=dev_mode)

        self._cfg.update(new_config.as_dict())

    def _get_value_verbose(self, key: str) -> Union[float, int, str, list, dict, Path, pd.Timestamp]:
        """Use this function internally to return attributes of the config that are mandatory"""
        if key not in self._cfg.keys():
            raise ValueError(f"{key} is not specified in the config (.yml).")
        elif self._cfg[key] is None:
            raise ValueError(f"{key} is mandatory but 'None' in the config.")
        else:
            return self._cfg[key]

    @staticmethod
    def _as_default_list(value: Any) -> list:
        if value is None:
            return []
        elif isinstance(value, list):
            return value
        else:
            return [value]

    @staticmethod
    def _as_default_dict(value: Any) -> dict:
        if value is None:
            return {}
        elif isinstance(value, dict):
            return value
        else:
            raise RuntimeError(f"Incompatible type {type(value)}. Expected `dict` or `None`.")

    @staticmethod
    def _check_cfg_keys(cfg: dict):
        """Checks the config for unknown keys. """
        property_names = [p for p in dir(Config) if isinstance(getattr(Config, p), property)]

        unknown_keys = [
            k for k in cfg.keys()
            if k not in property_names and k not in Config._deprecated_keys and k not in Config._metadata_keys
        ]
        if unknown_keys:
            raise ValueError(f'{unknown_keys} are not recognized config keys.')

    @staticmethod
    def _parse_config(cfg: dict) -> dict:
        for key, val in cfg.items():
            # convert all path strings to PosixPath objects
            if any([key.endswith(x) for x in ['_dir', '_path', '_file', '_files']]):
                if (val is not None) and (val != "None"):
                    if isinstance(val, list):
                        temp_list = []
                        for element in val:
                            temp_list.append(Path(element))
                        cfg[key] = temp_list
                    else:
                        cfg[key] = Path(val)
                else:
                    cfg[key] = None

            # convert Dates to pandas Datetime indexs
            elif key.endswith('_date'):
                if isinstance(val, list):
                    temp_list = []
                    for elem in val:
                        temp_list.append(pd.to_datetime(elem, format='%d/%m/%Y'))
                    cfg[key] = temp_list
                else:
                    cfg[key] = pd.to_datetime(val, format='%d/%m/%Y')

            else:
                pass

        # Check forecast sequence length.
        if cfg.get('forecast_seq_length'):
            if cfg['forecast_seq_length'] >= cfg['seq_length']:
                raise ValueError('Forecast sequence length must be < sequence length.')
            if cfg.get('forecast_overlap'):
                if cfg['forecast_overlap'] > cfg['forecast_seq_length']:
                    raise ValueError('Forecast overlap must be <= forecast_seq_length.')

        # Check autoregressive inputs.
        if 'autoregressive_inputs' in cfg:
            if len(cfg['autoregressive_inputs']) > 1:
                raise ValueError('Currently only one autoregressive input is supported.')
            if cfg['autoregressive_inputs'] and len(cfg['target_variables']) > 1:
                raise ValueError('Autoregressive models currently only support a single target variable.')
            if not cfg['autoregressive_inputs'][0].startswith(cfg['target_variables'][0]):
                raise ValueError('Autoregressive input must be a lagged version of the target variable.')

        # Add more config parsing if necessary
        return cfg

    @staticmethod
    def _read_and_parse_config(yml_path: Path):
        if yml_path.exists():
            with yml_path.open('r') as fp:
                yaml = YAML(typ="safe")
                cfg = yaml.load(fp)
        else:
            raise FileNotFoundError(yml_path)
        cfg = Config._parse_config(cfg)

        return cfg

    @property
    def additional_feature_files(self) -> List[Path]:
        return self._as_default_list(self._cfg.get("additional_feature_files", None))

    @property
    def allow_subsequent_nan_losses(self) -> int:
        return self._cfg.get("allow_subsequent_nan_losses", 0)

    @property
    def autoregressive_inputs(self) -> Union[List[str], Dict[str, List[str]]]:
        return self._as_default_list(self._cfg.get("autoregressive_inputs", []))

    @property
    def base_run_dir(self) -> Path:
        return self._get_value_verbose("base_run_dir")

    @base_run_dir.setter
    def base_run_dir(self, folder: Path):
        self._cfg["base_run_dir"] = folder

    @property
    def batch_size(self) -> int:
        return self._get_value_verbose("batch_size")

    @property
    def bidirectional_stacked_forecast_lstm(self) -> bool:
        return self._cfg.get("bidirectional_stacked_forecast_lstm", False)

    @property
    def cache_validation_data(self) -> bool:
        return self._cfg.get("cache_validation_data", True)

    @property
    def checkpoint_path(self) -> Path:
        return self._cfg.get("checkpoint_path", None)

    @property
    def clip_gradient_norm(self) -> float:
        return self._cfg.get("clip_gradient_norm", None)

    @property
    def clip_targets_to_zero(self) -> List[str]:
        return self._as_default_list(self._cfg.get("clip_targets_to_zero", []))

    @property
    def continue_from_epoch(self) -> int:
        return self._cfg.get("continue_from_epoch", None)

    @property
    def custom_normalization(self) -> dict:
        return self._as_default_dict(self._cfg.get("custom_normalization", {}))

    @property
    def data_dir(self) -> Path:
        return self._get_value_verbose("data_dir")

    @property
    def dataset(self) -> str:
        return self._get_value_verbose("dataset")

    @property
    def device(self) -> str:
        return self._cfg.get("device", None)

    @device.setter
    def device(self, device: str):
        if device == "cpu" or device.startswith("cuda:"):
            self._cfg["device"] = device
        else:
            raise ValueError("'device' must be either 'cpu' or a 'cuda:X', with 'X' being the GPU ID.")

    @property
    def duplicate_features(self) -> dict:
        duplicate_features = self._cfg.get("duplicate_features", {})
        if duplicate_features is None:
            return {}
        elif isinstance(duplicate_features, dict):
            return duplicate_features
        elif isinstance(duplicate_features, list):
            return {feature: 1 for feature in duplicate_features}
        elif isinstance(duplicate_features, str):
            return {duplicate_features: 1}
        else:
            raise RuntimeError(f"Unsupported type {type(duplicate_features)} for 'duplicate_features' argument.")

    @property
    def dynamic_conceptual_inputs(self) -> List[str]:
        return self._as_default_list(self._cfg.get("dynamic_conceptual_inputs", []))

    @property
    def warmup_period(self) -> int:
        return self._cfg.get("warmup_period", 0)

    @property
    def dynamic_inputs(self) -> Union[List[str], Dict[str, List[str]]]:
        return self._get_value_verbose("dynamic_inputs")

    @property
    def dynamics_embedding(self) -> dict:
        embedding_spec = self._cfg.get("dynamics_embedding", None)

        if embedding_spec is None:
            return None
        return self._get_embedding_spec(embedding_spec)

    @property
    def epochs(self) -> int:
        return self._get_value_verbose("epochs")

    @property
    def evolving_attributes(self) -> List[str]:
        if "evolving_attributes" in self._cfg.keys():
            return self._as_default_list(self._cfg["evolving_attributes"])
        elif "static_inputs" in self._cfg.keys():
            warnings.warn("'static_inputs' will be deprecated. Use 'evolving_attributes' in the future", FutureWarning)
            return self._as_default_list(self._cfg["static_inputs"])
        else:
            return []

    @property
    def experiment_name(self) -> str:
        if self._cfg.get("experiment_name", None) is None:
            return "run"
        else:
            return self._cfg["experiment_name"]

    @property
    def finetune_modules(self) -> Union[List[str], Dict[str, str]]:
        finetune_modules = self._cfg.get("finetune_modules", [])
        if finetune_modules is None:
            return []
        elif isinstance(finetune_modules, str):
            return [finetune_modules]
        elif isinstance(finetune_modules, dict) or isinstance(finetune_modules, list):
            return finetune_modules
        else:
            raise ValueError(f"Unknown data type {type(finetune_modules)} for 'finetune_modules' argument.")

    @property
    def forecast_network(self) -> dict:
        embedding_spec = self._cfg.get("forecast_network", None)

        if embedding_spec is None:
            return None
        return self._get_embedding_spec(embedding_spec)

    @property
    def forecast_hidden_size(self) -> int:
        return self._cfg.get("forecast_hidden_size", self.hidden_size)

    @property
    def forecast_inputs(self) -> List[str]:
        return self._cfg.get("forecast_inputs", [])

    @property
    def forecast_overlap(self) -> int:
        return self._cfg.get("forecast_overlap", None)

    @property
    def forecast_seq_length(self) -> int:
        return self._cfg.get("forecast_seq_length", None)

    @property
    def forcings(self) -> List[str]:
        return self._as_default_list(self._get_value_verbose("forcings"))

    @property
    def save_git_diff(self) -> bool:
        return self._cfg.get('save_git_diff', False)

    @property
    def state_handoff_network(self) -> dict:
        embedding_spec = self._cfg.get("state_handoff_network", None)

        if embedding_spec is None:
            return None
        return self._get_embedding_spec(embedding_spec)

    @property
    def head(self) -> str:
        if self.model == "mclstm":
            return ''
        else:
            return self._get_value_verbose("head")

    @property
    def hindcast_inputs(self) -> List[str]:
        return self._cfg.get("hindcast_inputs", [])

    @property
    def hidden_size(self) -> Union[int, Dict[str, int]]:
        return self._get_value_verbose("hidden_size")

    @property
    def hindcast_hidden_size(self) -> Union[int, Dict[str, int]]:
        return self._cfg.get("hindcast_hidden_size", self.hidden_size)

    @property
    def hydroatlas_attributes(self) -> List[str]:
        return self._as_default_list(self._cfg.get("hydroatlas_attributes", []))

    @property
    def img_log_dir(self) -> Path:
        return self._cfg.get("img_log_dir", None)

    @img_log_dir.setter
    def img_log_dir(self, folder: Path):
        self._cfg["img_log_dir"] = folder

    @property
    def initial_forget_bias(self) -> float:
        return self._cfg.get("initial_forget_bias", None)

    @property
    def is_continue_training(self) -> bool:
        return self._cfg.get("is_continue_training", False)

    @is_continue_training.setter
    def is_continue_training(self, flag: bool):
        self._cfg["is_continue_training"] = flag

    @property
    def is_finetuning(self) -> bool:
        return self._cfg.get("is_finetuning", False)

    @is_finetuning.setter
    def is_finetuning(self, flag: bool):
        self._cfg["is_finetuning"] = flag

    @property
    def lagged_features(self) -> dict:
        return self._as_default_dict(self._cfg.get("lagged_features", {}))

    @property
    def learning_rate(self) -> Dict[int, float]:
        if ("learning_rate" in self._cfg.keys()) and (self._cfg["learning_rate"] is not None):
            if isinstance(self._cfg["learning_rate"], float):
                return {0: self._cfg["learning_rate"]}
            elif isinstance(self._cfg["learning_rate"], dict):
                return self._cfg["learning_rate"]
            else:
                raise ValueError("Unsupported data type for learning rate. Use either dict (epoch to float) or float.")
        else:
            raise ValueError("No learning rate specified in the config (.yml).")

    @property
    def log_interval(self) -> int:
        return self._cfg.get("log_interval", 10)

    @property
    def log_n_figures(self) -> int:
        if (self._cfg.get("log_n_figures", None) is None) or (self._cfg["log_n_figures"] < 1):
            return 0
        else:
            return self._cfg["log_n_figures"]

    @property
    def log_tensorboard(self) -> bool:
        return self._cfg.get("log_tensorboard", True)

    @property
    def loss(self) -> str:
        return self._get_value_verbose("loss")

    @loss.setter
    def loss(self, loss: str):
        self._cfg["loss"] = loss

    @property
    def mass_inputs(self) -> List[str]:
        return self._as_default_list(self._cfg.get("mass_inputs", []))

    @property
    def max_updates_per_epoch(self) -> Optional[int]:
        return self._cfg.get("max_updates_per_epoch")

    @property
    def mc_dropout(self) -> bool:
        return self._cfg.get("mc_dropout", False)

    @property
    def metrics(self) -> Union[List[str], Dict[str, List[str]]]:
        return self._cfg.get("metrics", [])

    @metrics.setter
    def metrics(self, metrics: Union[str, List[str], Dict[str, List[str]]]):
        self._cfg["metrics"] = metrics

    @property
    def model(self) -> str:
        return self._get_value_verbose("model")

    @property
    def conceptual_model(self) -> str:
        return self._cfg.get("conceptual_model", "SHM")

    @property
    def n_distributions(self) -> int:
        return self._get_value_verbose("n_distributions")

    @property
    def n_samples(self) -> int:
        return self._get_value_verbose("n_samples")

    @property
    def n_taus(self) -> int:
        return self._get_value_verbose("n_taus")

    @property
    def negative_sample_handling(self) -> str:
        return self._cfg.get("negative_sample_handling", None)

    @property
    def negative_sample_max_retries(self) -> int:
        return self._get_value_verbose("negative_sample_max_retries")

    @property
    def no_loss_frequencies(self) -> list:
        return self._as_default_list(self._cfg.get("no_loss_frequencies", []))

    @property
    def num_workers(self) -> int:
        return self._cfg.get("num_workers", 0)

    @property
    def number_of_basins(self) -> int:
        return self._get_value_verbose("number_of_basins")

    @number_of_basins.setter
    def number_of_basins(self, num_basins: int):
        self._cfg["number_of_basins"] = num_basins

    @property
    def ode_method(self) -> str:
        return self._cfg.get("ode_method", "euler")

    @property
    def ode_num_unfolds(self) -> int:
        return self._cfg.get("ode_num_unfolds", 4)

    @property
    def ode_random_freq_lower_bound(self) -> str:
        return self._get_value_verbose("ode_random_freq_lower_bound")

    @property
    def optimizer(self) -> str:
        return self._get_value_verbose("optimizer")

    @property
    def output_activation(self) -> str:
        return self._cfg.get("output_activation", "linear")

    @property
    def output_dropout(self) -> float:
        return self._cfg.get("output_dropout", 0.0)

    @property
    def per_basin_test_periods_file(self) -> Path:
        return self._cfg.get("per_basin_test_periods_file", None)

    @property
    def per_basin_train_periods_file(self) -> Path:
        return self._cfg.get("per_basin_train_periods_file", None)

    @property
    def per_basin_validation_periods_file(self) -> Path:
        return self._cfg.get("per_basin_validation_periods_file", None)

    @property
    def predict_last_n(self) -> Union[int, Dict[str, int]]:
        return self._get_value_verbose("predict_last_n")

    @property
    def random_holdout_from_dynamic_features(self) -> Dict[str, float]:
        return self._as_default_dict(self._cfg.get("random_holdout_from_dynamic_features", {}))

    @property
    def rating_curve_file(self) -> Path:
        return self._get_value_verbose("rating_curve_file")

    @property
    def regularization(self) -> List[Union[str, Tuple[str, float]]]:
        return self._as_default_list(self._cfg.get("regularization", []))

    @property
    def run_dir(self) -> Path:
        return self._cfg.get("run_dir", None)

    @run_dir.setter
    def run_dir(self, folder: Path):
        self._cfg["run_dir"] = folder

    @property
    def save_train_data(self) -> bool:
        return self._cfg.get("save_train_data", False)

    @property
    def save_all_output(self) -> bool:
        return self._cfg.get('save_all_output', False)

    @property
    def save_validation_results(self) -> bool:
        return self._cfg.get("save_validation_results", False)

    @property
    def save_weights_every(self) -> int:
        return self._cfg.get("save_weights_every", 1)

    @property
    def seed(self) -> int:
        return self._cfg.get("seed", None)

    @property
    def transformer_nlayers(self) -> int:
        return self._get_value_verbose("transformer_nlayers")

    @property
    def transformer_positional_encoding_type(self) -> str:
        return self._get_value_verbose("transformer_positional_encoding_type")

    @property
    def transformer_dim_feedforward(self) -> int:
        return self._get_value_verbose("transformer_dim_feedforward")

    @property
    def transformer_positional_dropout(self) -> float:
        return self._get_value_verbose("transformer_positional_dropout")

    @property
    def transformer_dropout(self) -> float:
        return self._get_value_verbose("transformer_dropout")

    @property
    def transformer_nheads(self) -> int:
        return self._get_value_verbose("transformer_nheads")

    @seed.setter
    def seed(self, seed: int):
        if self._cfg.get("seed", None) is None:
            self._cfg["seed"] = seed
        else:
            raise RuntimeError("Seed was already specified and can't be replaced")

    @property
    def seq_length(self) -> Union[int, Dict[str, int]]:
        return self._get_value_verbose("seq_length")

    @property
    def shared_mtslstm(self) -> bool:
        return self._cfg.get("shared_mtslstm", False)

    @property
    def static_attributes(self) -> List[str]:
        if "static_attributes" in self._cfg.keys():
            return self._as_default_list(self._cfg["static_attributes"])
        elif "camels_attributes" in self._cfg.keys():
            warnings.warn("'camels_attributes' will be deprecated. Use 'static_attributes' in the future",
                          FutureWarning)
            return self._as_default_list(self._cfg["camels_attributes"])
        else:
            return []

    @property
    def statics_embedding(self) -> dict:
        embedding_spec = self._cfg.get("statics_embedding", None)

        if embedding_spec is None:
            return None
        return self._get_embedding_spec(embedding_spec)

    @property
    def target_loss_weights(self) -> List[float]:
        return self._cfg.get("target_loss_weights", None)

    @property
    def target_noise_std(self) -> float:
        if (self._cfg.get("target_noise_std", None) is None) or (self._cfg["target_noise_std"] == 0):
            return None
        else:
            return self._cfg["target_noise_std"]

    @property
    def target_variables(self) -> List[str]:
        if "target_variables" in self._cfg.keys():
            return self._cfg["target_variables"]
        elif "target_variable" in self._cfg.keys():
            warnings.warn("'target_variable' will be deprecated. Use 'target_variables' in the future", FutureWarning)
            return self._cfg["target_variable"]
        else:
            raise ValueError("No target variables ('target_variables') defined in the config.")

    @property
    def tau_down(self) -> float:
        return self._get_value_verbose("tau_down")

    @property
    def tau_up(self) -> float:
        return self._get_value_verbose("tau_up")

    @property
    def test_basin_file(self) -> Path:
        return self._get_value_verbose("test_basin_file")

    @property
    def test_end_date(self) -> pd.Timestamp:
        return self._get_value_verbose("test_end_date")

    @property
    def test_start_date(self) -> pd.Timestamp:
        return self._get_value_verbose("test_start_date")

    @property
    def timestep_counter(self) -> bool:
        return self._cfg.get("timestep_counter", False)

    @property
    def train_basin_file(self) -> Path:
        return self._get_value_verbose("train_basin_file")

    @property
    def train_data_file(self) -> Path:
        return self._cfg.get("train_data_file", None)

    @property
    def train_dir(self) -> Path:
        return self._cfg.get("train_dir", None)

    @train_dir.setter
    def train_dir(self, folder: Path):
        self._cfg["train_dir"] = folder

    @property
    def train_end_date(self) -> pd.Timestamp:
        return self._get_value_verbose("train_end_date")

    @property
    def train_start_date(self) -> pd.Timestamp:
        return self._get_value_verbose("train_start_date")

    @property
    def transfer_mtslstm_states(self) -> Dict[str, str]:
        return self._cfg.get("transfer_mtslstm_states", {'h': 'linear', 'c': 'linear'})

    @property
    def umal_extend_batch(self) -> bool:
        return self._cfg.get("umal_extend_batch", False)

    @property
    def use_basin_id_encoding(self) -> bool:
        return self._cfg.get("use_basin_id_encoding", False)

    @property
    def use_frequencies(self) -> List[str]:
        return self._as_default_list(self._cfg.get("use_frequencies", []))

    @property
    def validate_every(self) -> int:
        if (self._cfg.get("validate_every", None) is None) or (self._cfg["validate_every"] < 1):
            return None
        else:
            return self._cfg["validate_every"]

    @property
    def validate_n_random_basins(self) -> int:
        if (self._cfg.get("validate_n_random_basins", None) is None) or (self._cfg["validate_n_random_basins"] < 1):
            return 0
        else:
            return self._cfg["validate_n_random_basins"]

    @validate_n_random_basins.setter
    def validate_n_random_basins(self, n_basins: int):
        self._cfg["validate_n_random_basins"] = n_basins

    @property
    def validation_basin_file(self) -> Path:
        return self._get_value_verbose("validation_basin_file")

    @property
    def validation_end_date(self) -> pd.Timestamp:
        return self._get_value_verbose("validation_end_date")

    @property
    def validation_start_date(self) -> pd.Timestamp:
        return self._get_value_verbose("validation_start_date")

    @property
    def verbose(self) -> int:
        """Defines level of verbosity.

        0: Only log info messages, don't show progress bars
        1: Log info messages and show progress bars

        Returns
        -------
        int
            Level of verbosity.
        """
        return self._cfg.get("verbose", 1)

    def _get_embedding_spec(self, embedding_spec: dict) -> dict:
        if isinstance(embedding_spec, bool) and embedding_spec:  #
            msg = [
                "The semantics of 'dynamics/statics_embedding' have changed, and the associated arguments "
                "'embedding_hiddens/activation/dropout' are deprecated. The old specifications may no longer work in "
                "the future. Specify embeddings as a dict in dynamics/statics_embedding instead."
            ]
            warnings.warn(" ".join(msg), FutureWarning)
            return {
                'type': 'fc',
                'hiddens': self._as_default_list(self._cfg.get("embedding_hiddens", [])),
                'activation': self._cfg.get("embedding_activation", "tanh"),
                'dropout': self._cfg.get("embedding_dropout", 0.0)
            }

        return {
            'type': embedding_spec.get('type', 'fc'),
            'hiddens': self._as_default_list(embedding_spec.get('hiddens', [])),
            'activation': embedding_spec.get('activation', 'tanh'),
            'dropout': embedding_spec.get('dropout', 0.0)
        }


def create_random_name():
    adjectives = ('white', 'black', 'green', 'golden', 'modern', 'lazy', 'great', 'meandering', 'nervous', 'demanding',
                  'relaxed', 'dashing', 'clever', 'brave', 'charming')
    nouns = ('mississippi', 'amazon', 'nile', 'yangtze', 'yellow', 'congo', 'mekong', 'otter', 'frog', 'trout',
             'beaver', 'eel', 'catfish', 'salmon')

    rng = random.Random(datetime.now().timestamp())  # use system time as random seed
    return rng.choice(adjectives) + '-' + rng.choice(nouns)



"""Unit tests for datautils functions. """
import pandas as pd
import pytest

from neuralhydrology.datautils.utils import get_frequency_factor, infer_frequency, sort_frequencies


def test_sort_frequencies():
    """Test the sorting of frequencies. """
    assert sort_frequencies(['1D', '1H', '2D', '3H']) == ['2D', '1D', '3H', '1H']
    assert sort_frequencies(['1M', '1Y']) == ['1Y', '1M']
    assert sort_frequencies(['1D', '48H']) == ['48H', '1D']
    assert sort_frequencies(['1D']) == ['1D']
    assert sort_frequencies([]) == []

    pytest.raises(ValueError, sort_frequencies, ['1D', '1XYZ'])  # not a frequency


def test_infer_frequency():
    """Test the logic to infer frequencies. """
    assert infer_frequency(pd.date_range('2000-01-01', '2000-01-10', freq='D')) == '1D'
    assert infer_frequency(pd.date_range('2000-01-01', '2000-01-10', freq='48H')) == '2D'
    assert infer_frequency(pd.date_range('2000-01-01', '2000-03-01', freq='W-MON')) == '7D'

    # just a single date
    pytest.raises(ValueError, infer_frequency, [pd.to_datetime('2000-01-01')])
    # frequency of zero
    pytest.raises(ValueError, infer_frequency,
                  [pd.to_datetime('2000-01-01'),
                   pd.to_datetime('2000-01-01'),
                   pd.to_datetime('2000-01-01')])
    # irregular dates
    pytest.raises(ValueError, infer_frequency,
                  [pd.to_datetime('2000-01-01'),
                   pd.to_datetime('2000-01-02'),
                   pd.to_datetime('2000-01-04')])


def test_get_frequency_factor():
    """Test the logic that calculates the ratio between two frequencies. """
    assert get_frequency_factor('1H', '1H') == 1
    assert get_frequency_factor('1A', '1Y') == 1
    assert get_frequency_factor('1Y', '4Q') == 1
    assert get_frequency_factor('1H', '1D') == 1 / 24
    assert get_frequency_factor('1D', '1H') == 24
    assert get_frequency_factor('2D', '12H') == 4
    assert get_frequency_factor('1W', '1D') == 7
    assert get_frequency_factor('1W-MON', '1D') == 7
    assert get_frequency_factor('1Y', '1M') == 12
    assert get_frequency_factor('0D', '0H') == 1

    pytest.raises(ValueError, get_frequency_factor, '1YS', '1M')  # year-start vs. month-end
    pytest.raises(ValueError, get_frequency_factor, '1Q', '1W')  # quarter vs. week
    pytest.raises(ValueError, get_frequency_factor, '1XYZ', '1D')  # not a frequency
    pytest.raises(ValueError, get_frequency_factor, '1Y', '1D')  # disallowed because to_timedelta('1Y') is deprecated
    pytest.raises(ValueError, get_frequency_factor, '1M', '1D')  # disallowed because to_timedelta('1M') is deprecated
    pytest.raises(NotImplementedError, get_frequency_factor, '-1D', '1H')  # we should never need negative frequencies



from typing import List

import numpy as np
import torch
import torch.nn as nn


class FC(nn.Module):
    """Auxiliary class to build (multi-layer) fully-connected networks.

    This class is used to build fully-connected embedding networks for static and/or dynamic input data.
    Use the config argument `statics/dynamics_embedding` to specify the architecture of the embedding network. See the
    `InputLayer` class on how to specify the exact embedding architecture.

    Parameters
    ----------
    input_size : int
        Number of input features.
    hidden_sizes : List[int]
        Size of the hidden and output layers.
    activation : str, optional
        Activation function for intermediate layers, default tanh.
    dropout : float, optional
        Dropout rate in intermediate layers.
    """

    def __init__(self, input_size: int, hidden_sizes: List[int], activation: str = 'tanh', dropout: float = 0.0):
        super(FC, self).__init__()

        if len(hidden_sizes) == 0:
            raise ValueError('hidden_sizes must at least have one entry to create a fully-connected net.')

        self.output_size = hidden_sizes[-1]
        hidden_sizes = hidden_sizes[:-1]

        activation = self._get_activation(activation)

        # create network
        layers = []
        if hidden_sizes:
            for i, hidden_size in enumerate(hidden_sizes):
                if i == 0:
                    layers.append(nn.Linear(input_size, hidden_size))
                else:
                    layers.append(nn.Linear(hidden_sizes[i - 1], hidden_size))

                layers.append(activation)
                layers.append(nn.Dropout(p=dropout))

            layers.append(nn.Linear(hidden_size, self.output_size))
        else:
            layers.append(nn.Linear(input_size, self.output_size))

        self.net = nn.Sequential(*layers)
        self._reset_parameters()

    def _get_activation(self, name: str) -> nn.Module:
        if name.lower() == "tanh":
            activation = nn.Tanh()
        elif name.lower() == "sigmoid":
            activation = nn.Sigmoid()
        elif name.lower() == "linear":
            activation = nn.Identity()
        else:
            raise NotImplementedError(f"{name} currently not supported as activation in this class")
        return activation

    def _reset_parameters(self):
        """Special initialization of certain model weights."""
        for layer in self.net:
            if isinstance(layer, nn.modules.linear.Linear):
                n_in = layer.weight.shape[1]
                gain = np.sqrt(3 / n_in)
                nn.init.uniform_(layer.weight, -gain, gain)
                nn.init.constant_(layer.bias, val=0)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Perform a forward pass on the FC model.

        Parameters
        ----------
        x : torch.Tensor
            Input data of shape [any, any, input size]

        Returns
        -------
        torch.Tensor
            Embedded inputs of shape [any, any, output_size], where 'output_size' is the size of the last network layer.
        """
        return self.net(x)



from typing import Dict

import torch

from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.utils.config import Config


class TemplateModel(BaseModel):
    """Initialize the model

    Each model constructor receives the config instance as its only input. From this, the entire model has to be
    created (with potential use of other modules, such as FC from fc.py). During training and inference, the forward
    method of this class will be called with the model inputs and has to return the predictions.

    Each Model inherits from the BaseModel, which implements some universal functionality. The basemodel also 
    defines the output_size, which can be used here as a given attribute (self.output_size).

    To be generally useable within this codebase, the output layer should not be implemented in this Module,
    but rather using the get_head() function from neuralhydrology.modelzoo.head.

    Parameters
    ----------
    cfg : Config
        Configuration of the run, read from the config file with some additional keys (such as number of basins).
    """
    # specify submodules of the model that can later be used for finetuning. Names must match class attributes
    module_parts = []

    def __init__(self, cfg: Config):

        super(TemplateModel, self).__init__(cfg=cfg)

        ###########################
        # Create model parts here #
        ###########################

    def forward(self, data: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Forward pass through the model

                By convention, each forward pass has to accept a dict of input tensors. Usually, this dict contains
                'x_d' and, possibly, x_s and x_one_hot. If x_d and x_s are available at multiple frequencies,
                the keys 'x_d' and 'x_s' have frequency suffixes such as 'x_d_1H' for hourly data.
                Furthermore, by definition, each model has to return a dict containing the network predictions in
                'y_hat', potentially in addition to other dictionary keys. LSTM-based models should stick to the
                convention to return (at least) the following three tensors: y_hat, h_n, c_n (or, in the multi-
                frequency case, y_hat_1H, y_hat_1D, etc.).

                Parameters
                ----------
                data : Dict[str, torch.Tensor]
                     Dictionary with tensors
                        - x_d of shape [batch size, sequence length, features] containing the dynamic input data.
                        - x_s of shape [batch size, features] containing static input features. These are the
                            concatenation of what is defined in the config under static_attributes and evolving_attributes.
                            In case not a single camels attribute or static input feature is defined in the config,
                            x_s will not be present.
                        - x_one_hot of shape [batch size, number of basins] containing the one hot encoding of the
                            basins. In case 'use_basin_id_encoding' is set to False in the config, x_one_hot will
                            not be present.
                            
                        Note: If the input data are available at multiple frequencies (via use_frequencies), each input
                            tensor will have a suffix "_{freq}" indicating the tensor's frequency.

                Returns
                -------
                The network prediction has to be returned under the dictionary key 'y_hat' (or, if multiple frequencies
                are predicted, 'y_hat_{freq}'. Furthermore, make sure to return predictions for each time step, even if
                you want to train sequence-to-one. Which predictions are used for training the network is controlled in
                the train_epoch() function in neuralhydrology/training/basetrainer.py. Other return values should be the
                hidden states as 'h_n' and cell states 'c_n'. Further return values are possible.
                """
        ###############################
        # Implement forward pass here #
        ###############################
        pass




from typing import Dict

import torch
import torch.nn as nn

from neuralhydrology.modelzoo.inputlayer import InputLayer
from neuralhydrology.modelzoo.head import get_head
from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.utils.config import Config
from neuralhydrology.modelzoo.fc import FC


class MultiHeadForecastLSTM(BaseModel):
    """A forecasting model that does not roll out over the forecast horizon.
    
    This is a forecasting model that runs a sequential (LSTM) model up to the forecast issue time, 
    and then directly predicts a sequence of forecast timesteps without using a recurrent rollout.
    Prediction is done with a custom ``FC`` (fully connected) layer, which can include depth.

    Do not use this model with ``forecast_overlap`` > 0.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    
    Raises
    ------
    ValueError if forecast_overlap > 0.
    ValueError if a forecast_network is not specified.
    """
    # specify submodules of the model that can later be used for finetuning. Names must match class attributes
    module_parts = [
        'forecast_mebedding_net',
        'hindcast_embedding_net',
        'hindcast_lstm',
        'forecast_network',
        'hindcast_head',
        'forecast_head'
    ]

    def __init__(self, cfg: Config):
        super(MultiHeadForecastLSTM, self).__init__(cfg=cfg)

        if cfg.forecast_overlap:
            raise ValueError('Forecast overlap cannot be set for a multi-head forecasting model. '
                             'Please set to None or remove from config file.')

        self.forecast_embedding_net = InputLayer(cfg, embedding_type='forecast')
        self.hindcast_embedding_net = InputLayer(cfg, embedding_type='hindcast')

        self.hindcast_lstm = nn.LSTM(
            input_size=self.hindcast_embedding_net.output_size,
            hidden_size=cfg.hidden_size
        )

        self.dropout = nn.Dropout(p=cfg.output_dropout)

        if not cfg.forecast_network:
            raise ValueError('The multihead forecast model requires a forecast network specified in the config file.')

        input_size = self.forecast_embedding_net.output_size*cfg.forecast_seq_length + cfg.hidden_size
        forecast_network_output_size = cfg.forecast_network['hiddens'][-1] * cfg.forecast_seq_length
        self.forecast_network = FC(
            input_size=input_size,
            hidden_sizes=cfg.forecast_network['hiddens'][:-1] + [forecast_network_output_size],
            activation=cfg.forecast_network['activation'],
            dropout=cfg.forecast_network['dropout']
        )

        self.hindcast_head = get_head(cfg=cfg, n_in=cfg.hidden_size, n_out=self.output_size)
        self.forecast_head = get_head(cfg=cfg, n_in=cfg.forecast_network['hiddens'][-1], n_out=self.output_size)

        self._reset_parameters()

    def _reset_parameters(self):
        """Special initialization of certain model weights."""
        if self.cfg.initial_forget_bias is not None:
            self.hindcast_lstm.bias_hh_l0.data[self.cfg.hidden_size:2 * self.cfg.hidden_size] = self.cfg.initial_forget_bias

    def forward(self, data: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform a forward pass on the MultiheadForecastLSTM model.
        
        Parameters
        ----------
        data : Dict[str, torch.Tensor]
            Dictionary, containing input features as key-value pairs.
        
        Returns
        -------
        Dict[str, torch.Tensor]
            Model outputs and intermediate states as a dictionary.
                - lstm_output_hindcast: Output sequence from the hindcast LSTM.
                - output_forecast: Predictions (before head layer) from the forecast period.
                - h_n_hindcast: Final hidden state of the hindcast model.
                - c_n_hindcast: Final cell state of the hindcast model.
                - y_hat: Predictions over the sequence from the head layer.
        """
        # possibly pass dynamic and static inputs through embedding layers, then concatenate them
        x_h = self.hindcast_embedding_net(data)
        x_f = self.forecast_embedding_net(data)

        # run hindcast part of the lstm
        lstm_output_hindcast, (h_n_hindcast, c_n_hindcast) = self.hindcast_lstm(input=x_h)
        lstm_output_hindcast = lstm_output_hindcast.transpose(0, 1)
        output_hindcast = self.hindcast_head(self.dropout(lstm_output_hindcast))

        # reshape to [batch_size, seq, n_hiddens]
        h_n_hindcast = h_n_hindcast.transpose(0, 1)
        c_n_hindcast = c_n_hindcast.transpose(0, 1)

        # run forecast heads
        batch_size = x_f.shape[1]
        x_f = x_f.transpose(0, 1).contiguous()
        x = torch.cat([h_n_hindcast.squeeze(dim=1), x_f.view(batch_size, -1)], dim=-1)
        x = self.forecast_network(x)
        x = x.view(batch_size, self.cfg.forecast_seq_length, -1)
        output_forecast = self.forecast_head(self.dropout(x))

        # start an output dictionary
        pred = {key: torch.cat([output_hindcast[key], output_forecast[key]], dim=1) for key in output_hindcast}

        pred.update(
            {
                'lstm_output_hindcast': lstm_output_hindcast,
                'output_forecast': output_forecast,

                'h_n_hindcast': h_n_hindcast,
                'c_n_hindcast': c_n_hindcast,
            }
        )

        return pred



from pathlib import Path
from typing import List, Dict, Union

import numpy as np
import pandas as pd
import xarray

from neuralhydrology.datasetzoo.basedataset import BaseDataset
from neuralhydrology.datautils import utils
from neuralhydrology.utils.config import Config

_SUBDATASET_TO_DIRECTORY = {
    'lamah_a': 'A_basins_total_upstrm',
    'lamah_b': 'B_basins_intermediate_all',
    'lamah_c': 'C_basins_intermediate_lowimp'
}


class LamaH(BaseDataset):
    """Data set class for the LamaH-CE dataset by [#]_.
    
    The LamaH-CE dataset consists of three different catchment delineations, each with dedicated forcing time series and
    catchment attributes. These subdatasets are stored in the folder 'A_basins_total_upstrm', 
    'B_basins_intermediate_all', and 'C_basins_intermediate_lowimp'. The different datasets can be used by setting the
    config argument `dataset` to `lamah_a`, `lamah_b` or `lamah_c` for 'A_basins_total_upstrm', 'B_basins_intermediate_all',
    or 'C_basins_intermediate_lowimp', respectively. Furthermore, if you download the full dataset, each of these 
    subdatasets, as well as the streamflow data, comes at hourly and daily resolution. Based on the config argument 
    `use_frequencies` this dataset class will load daily data (for daily resolutions or lower), or hourly data (for all
    temporal resolutions higher than daily). If nothing is specified in `use_frequencies`, daily data is loaded by
    default. Also note: discharge data in the LamaH dataset is provided in m3s-1. This dataset class will transform
    discharge into mmd-1 (for daily data) or mmh-1 (for hourly data), using the 'area_gov' provided in the attributes
    file.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    is_train : bool 
        Defines if the dataset is used for training or evaluating. If True (training), means/stds for each feature
        are computed and stored to the run directory. If one-hot encoding is used, the mapping for the one-hot encoding 
        is created and also stored to disk. If False, a `scaler` input is expected and similarly the `id_to_int` input
        if one-hot encoding is used. 
    period : {'train', 'validation', 'test'}
        Defines the period for which the data will be loaded
    basin : str, optional
        If passed, the data for only this basin will be loaded. Otherwise the basin(s) are read from the appropriate
        basin file, corresponding to the `period`.
    additional_features : List[Dict[str, pd.DataFrame]], optional
        List of dictionaries, mapping from a basin id to a pandas DataFrame. This DataFrame will be added to the data
        loaded from the dataset, and all columns are available as 'dynamic_inputs', 'evolving_attributes' and
        'target_variables'
    id_to_int : Dict[str, int], optional
        If the config argument 'use_basin_id_encoding' is True in the config and period is either 'validation' or 
        'test', this input is required. It is a dictionary, mapping from basin id to an integer (the one-hot encoding).
    scaler : Dict[str, Union[pd.Series, xarray.DataArray]], optional
        If period is either 'validation' or 'test', this input is required. It contains the centering and scaling
        for each feature and is stored to the run directory during training (train_data/train_data_scaler.yml).

    References
    ----------
    .. [#]  Klingler, C., Schulz, K., and Herrnegger, M.: LamaH-CE: LArge-SaMple DAta for Hydrology and Environmental 
        Sciences for Central Europe, Earth Syst. Sci. Data, 13, 4529-4565, https://doi.org/10.5194/essd-13-4529-2021, 
        2021. 
    """

    def __init__(self,
                 cfg: Config,
                 is_train: bool,
                 period: str,
                 basin: str = None,
                 additional_features: List[Dict[str, pd.DataFrame]] = [],
                 id_to_int: Dict[str, int] = {},
                 scaler: Dict[str, Union[pd.Series, xarray.DataArray]] = {}):
        # Discharge is provided in m3/s in the data set as 'Qobs [m3/s]'. We allow to use 'Qobs [mm/h]' or 'Qobs [mm/d]'
        # in the config and in this case will normalize on the fly. For that, we need the basin area
        self._all_variables = self._get_list_of_all_variables(cfg)
        if any([f.startswith("qobs") for f in self._all_variables]):
            df = load_lamah_attributes(cfg.data_dir, sub_dataset=cfg.dataset)
            self._basin_area = df["area_gov"]

        # initialize parent class
        super(LamaH, self).__init__(cfg=cfg,
                                    is_train=is_train,
                                    period=period,
                                    basin=basin,
                                    additional_features=additional_features,
                                    id_to_int=id_to_int,
                                    scaler=scaler)

    @staticmethod
    def _get_list_of_all_variables(cfg: Config) -> List[str]:
        all_variables = []
        if isinstance(cfg.target_variables, dict):
            for val in cfg.target_variables.values():
                all_variables = all_variables + val
        else:
            all_variables = all_variables + cfg.target_variables
        if isinstance(cfg.dynamic_inputs, dict):
            for val in cfg.dynamic_inputs.values():
                all_variables = all_variables + val
        else:
            all_variables = all_variables + cfg.dynamic_inputs
        return all_variables

    def _load_basin_data(self, basin: str) -> pd.DataFrame:
        """Load input and output data from text files."""

        # Determine if hourly or daily data should be loaded, by default daily.
        temporal_resolution = '1D'
        if self.cfg.use_frequencies:
            if any([utils.compare_frequencies(freq, '1D') == 1 for freq in self.cfg.use_frequencies]):
                temporal_resolution = '1H'

        df = load_lamah_forcing(data_dir=self.cfg.data_dir,
                                basin=basin,
                                sub_dataset=self.cfg.dataset,
                                temporal_resolution=temporal_resolution)
        if any([f.startswith('qobs') for f in self._all_variables]):
            # We normalize discharge here to make use of the cached basin areas.
            discharge = load_lamah_discharge(data_dir=self.cfg.data_dir,
                                             basin=basin,
                                             temporal_resolution=temporal_resolution,
                                             normalize_discharge=False)
            df["qobs"] = _normalize_discharge(ser=discharge["qobs"],
                                              area=self._basin_area.loc[basin],
                                              temporal_resolution=temporal_resolution)

        return df

    def _load_attributes(self) -> pd.DataFrame:
        """Load static catchment attributes."""
        return load_lamah_attributes(self.cfg.data_dir, sub_dataset=self.cfg.dataset, basins=self.basins)


def load_lamah_forcing(data_dir: Path, basin: str, sub_dataset: str, temporal_resolution: str = '1D') -> pd.DataFrame:
    """Load forcing data of the LamaH data set.

    Parameters
    ----------
    data_dir : Path
        Path to the LamaH directory. 
    basin : str
        Basin identifier number as string.
    sub_dataset: str
        One of {'lamah_a', 'lamah_b', 'lamah_c'}, defining which of the three catchment delinations/sub-datasets 
        (A_basins_total_upstrm, B_basins_intermediate_all, or C_basins_intermediate_lowimp) will be loaded.
    temporal_resolution: str, optional
        Defines if either daily ('1D', default) or hourly ('1H') timeseries data will be loaded.

    Returns
    -------
    pd.DataFrame
        Time-indexed DataFrame, containing the forcings data.

    Raises
    ------
    ValueError
        If 'sub_dataset' is not one of {'lamah_a', 'lamah_b', 'lamah_c'}.
    ValueError
        If 'temporal_resolution' is not one of ['1H', '1D'].
    """
    if sub_dataset not in _SUBDATASET_TO_DIRECTORY:
        raise ValueError(
            f"{sub_dataset} is not a valid choice for 'sub_dataset'. Must be one of {_SUBDATASET_TO_DIRECTORY.keys()}.")

    if temporal_resolution not in ['1D', '1H']:
        raise ValueError(
            f"{temporal_resolution} is not a valid choice for 'temporal_resolution'. Must be one of '1H', '1D'.")

    temporal_resolution_directory = 'daily' if temporal_resolution == '1D' else 'hourly'

    # Load forcing data
    forcing_dir = data_dir / _SUBDATASET_TO_DIRECTORY[sub_dataset] / '2_timeseries' / temporal_resolution_directory
    return _load_lamah_timeseries_csv_file(forcing_dir / f"ID_{basin}.csv", temporal_resolution)


def load_lamah_discharge(data_dir: Path,
                         basin: str,
                         temporal_resolution: str = '1D',
                         normalize_discharge: bool = False) -> pd.DataFrame:
    """Load discharge data of the LamaH data set.

    Parameters
    ----------
    data_dir : Path
        Path to the LamaH directory. 
    basin : str
        Basin identifier number as string.
    temporal_resolution: str, optional
        Defines if either daily ('1D', default) or hourly ('1H') timeseries data will be loaded.
    normalize_discharge: bool, optional
        If true, normalizes discharge data by basin area, using the 'area_gov' attribute from attribute file.

    Returns
    -------
    pd.DataFrame
        Time-indexed DataFrame, containing the forcings data.

    Raises
    ------
    ValueError
        If 'temporal_resolution' is not one of ['1H', '1D'].
    """
    if temporal_resolution not in ['1D', '1H']:
        raise ValueError(
            f"{temporal_resolution} is not a valid choice for 'temporal_resolution'. Must be one of '1H', '1D'.")

    temporal_resolution_directory = 'daily' if temporal_resolution == '1D' else 'hourly'
    streamflow_dir = data_dir / 'D_gauges' / '2_timeseries' / temporal_resolution_directory
    df = _load_lamah_timeseries_csv_file(streamflow_dir / f"ID_{basin}.csv", temporal_resolution)

    # Replace missing discharge values (indicated by -999) to NaN
    df.loc[df["qobs"] < 0, "qobs"] = np.nan

    # If normalize_discharge is True, load attributes to extract upstream area.
    if normalize_discharge:
        attributes = load_lamah_attributes(data_dir, sub_dataset='lamah_a', basins=[basin])
        area = attributes.loc[basin, 'area_gov']
        df["qobs"] = _normalize_discharge(df["qobs"], area, temporal_resolution)

    return df


def load_lamah_attributes(data_dir: Path, sub_dataset: str, basins: List[str] = []) -> pd.DataFrame:
    """Load LamaH catchment attributes.

    Parameters
    ----------
    data_dir : Path
        Path to the LamaH-CE directory.
    sub_dataset: str
        One of {'lamah_a', 'lamah_b', 'lamah_c'}, defining which of the three catchment delinations/sub-datasets 
        (A_basins_total_upstrm, B_basins_intermediate_all, or C_basins_intermediate_lowimp) will be loaded.
    basins : List[str], optional
        If passed, return only attributes for the basins specified in this list. Otherwise, the attributes of all basins
        are returned.

    Returns
    -------
    pd.DataFrame
        Basin-indexed DataFrame, containing the attributes of the sub-dataset as well as the gauge attributes.

    Raises
    ------
    ValueError
        If any of the basin ids is not in the basin index.
    """
    # Load catchment attributes for sub_dataset.
    file_path = data_dir / _SUBDATASET_TO_DIRECTORY[sub_dataset] / "1_attributes" / "Catchment_attributes.csv"
    df_catchment = _load_lamah_attribute_csv_file(file_path)

    # Load gauge attributes
    file_path = data_dir / 'D_gauges' / "1_attributes" / "Gauge_attributes.csv"
    df_gauge = _load_lamah_attribute_csv_file(file_path)

    # Combine both DatFrames.
    df = pd.concat([df_catchment, df_gauge], axis=1)

    if basins:
        if any(b not in df.index for b in basins):
            raise ValueError('Some basins are missing static attributes.')
        df = df.loc[basins]

    return df


def _load_lamah_timeseries_csv_file(filepath: Path, temporal_resolution: str) -> pd.DataFrame:
    """Helper function to load lamah data into time indexed dataframe."""
    df = pd.read_csv(filepath, sep=';', dtype={'YYYY': str, 'MM': str, 'DD': str})
    df["date"] = pd.to_datetime(df.YYYY.map(str) + "/" + df.MM.map(str) + "/" + df.DD.map(str), format="%Y/%m/%d")
    if temporal_resolution == "1D":
        df = df.drop(['YYYY', 'MM', 'DD'], axis=1)
    else:
        df["date"] = df["date"] + pd.to_timedelta(df['hh'], unit='h')
        df = df.drop(['YYYY', 'MM', 'DD', 'hh', 'mm'], axis=1)
    return df.set_index('date')


def _load_lamah_attribute_csv_file(file_path: Path) -> pd.DataFrame:
    """Helper function to load lamah attribute files into basin indexed dataframes."""
    df = pd.read_csv(file_path, sep=";", dtype={'ID': str})
    df = df.set_index("ID")
    df.index.name = "gauge_id"
    return df


def _normalize_discharge(ser: pd.Series, area: float, temporal_resolution: str) -> pd.Series:
    """Helper function to normalize discharge data by basin area"""
    if temporal_resolution == "1H":
        return ser / (area * 1e6) * 1000 * 3600
    else:
        return ser / (area * 1e6) * 1000 * 86400


import logging
from typing import Dict, Optional, Union, Tuple

import torch
import torch.nn as nn

from neuralhydrology.modelzoo.fc import FC
from neuralhydrology.utils.config import Config

LOGGER = logging.getLogger(__name__)

_EMBEDDING_TYPES = ['full_model', 'hindcast', 'forecast']

class InputLayer(nn.Module):
    """Input layer to preprocess static and dynamic inputs.

    This module provides optional embedding of dynamic and static inputs. If ``dynamic_embeddings`` or
    ``static_embeddings`` are specified as dictionaries in the config, a fully-connected embedding network will be
    prepended to the timeseries model. The dictionaries have the following keys:

    - ``type`` (default 'fc'): Type of the embedding net. Currently, only 'fc' for fully-connected net is supported.
    - ``hiddens``: List of integers that define the number of neurons per layer in the fully connected network.
      The last number is the number of output neurons. Must have at least length one.
    - ``activation`` (default 'tanh'): activation function of the network. Supported values are 'tanh', 'sigmoid',
      'linear'. The activation function is not applied to the output neurons, which always have a linear activation
      function. An activation function for the output neurons has to be applied in the main model class.
    - ``dropout`` (default 0.0): Dropout rate applied to the embedding network.

    Note that this module does not support multi-frequency runs.

    Parameters
    ----------
    cfg : Config
        The run configuration
    """

    def __init__(self, cfg: Config, embedding_type: str = 'full_model'):
        super(InputLayer, self).__init__()

        if embedding_type not in _EMBEDDING_TYPES:
            raise ValueError(
                f'Embedding type {embedding_type} is not recognized. '
                f'Must be one of: {_EMBEDDING_TYPES}.'
            )
        self.embedding_type = embedding_type
        if embedding_type == 'full_model':
            dynamic_inputs = cfg.dynamic_inputs
        elif embedding_type == 'forecast':
            dynamic_inputs = cfg.forecast_inputs
        elif embedding_type == 'hindcast':
            dynamic_inputs = cfg.hindcast_inputs
            
        if isinstance(dynamic_inputs, dict):
            frequencies = list(dynamic_inputs.keys())
            if len(frequencies) > 1:
                raise ValueError('InputLayer only supports single-frequency data')
            dynamics_input_size = len(dynamic_inputs[frequencies[0]])
        else:
            dynamics_input_size = len(dynamic_inputs)

        self._num_autoregression_inputs = 0
        if cfg.autoregressive_inputs:
            self._num_autoregression_inputs = len(cfg.autoregressive_inputs)

        statics_input_size = len(cfg.static_attributes + cfg.hydroatlas_attributes + cfg.evolving_attributes)
        if cfg.use_basin_id_encoding:
            statics_input_size += cfg.number_of_basins

        self.statics_embedding, self.statics_output_size = \
            self._get_embedding_net(cfg.statics_embedding, statics_input_size, 'statics')
        self.dynamics_embedding, self.dynamics_output_size = \
            self._get_embedding_net(cfg.dynamics_embedding, dynamics_input_size, 'dynamics')

        if cfg.statics_embedding is None:
            self.statics_embedding_p_dropout = 0.0  # if net has no statics dropout we treat is as zero
        else:
            self.statics_embedding_p_dropout = cfg.statics_embedding['dropout']
        if cfg.dynamics_embedding is None:
            self.dynamics_embedding_p_dropout = 0.0  # if net has no dynamics dropout we treat is as zero
        else:
            self.dynamics_embedding_p_dropout = cfg.dynamics_embedding['dropout']

        self.output_size = self.dynamics_output_size + self.statics_output_size + self._num_autoregression_inputs
        if cfg.head.lower() == "umal":
            self.output_size += 1

    @staticmethod
    def _get_embedding_net(embedding_spec: Optional[dict], input_size: int, purpose: str) -> Tuple[nn.Module, int]:
        """Get an embedding net following the passed specifications.

        If the `embedding_spec` is None, the returned embedding net will be the identity function.

        Parameters
        ----------
        embedding_spec : Optional[dict]
            Specification of the embedding net from the run configuration or None.
        input_size : int
            Size of the inputs into the embedding network.
        purpose : str
            Purpose of the embedding network, used for error messages.

        Returns
        -------
        Tuple[nn.Module, int]
            The embedding net and its output size.
        """
        if embedding_spec is None:
            return nn.Identity(), input_size

        if input_size == 0:
            raise ValueError(f'Cannot create {purpose} embedding layer with input size 0')

        emb_type = embedding_spec['type'].lower()
        if emb_type != 'fc':
            raise ValueError(f'{purpose} embedding type {emb_type} not supported.')

        hiddens = embedding_spec['hiddens']
        if len(hiddens) == 0:
            raise ValueError(f'{purpose} embedding "hiddens" must be a list of hidden sizes with at least one entry')

        dropout = embedding_spec['dropout']
        activation = embedding_spec['activation']

        emb_net = FC(input_size=input_size, hidden_sizes=hiddens, activation=activation, dropout=dropout)
        return emb_net, emb_net.output_size

    def forward(self, data: Dict[str, torch.Tensor], concatenate_output: bool = True) \
            -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Perform a forward pass on the input layer.

        Parameters
        ----------
        data : Dict[str, torch.Tensor]
            The input data.
        concatenate_output : bool, optional
            If True (default), the forward method will concatenate the static inputs to each dynamic time step.
            If False, the forward method will return a tuple of (dynamic, static) inputs.

        Returns
        -------
        Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]
            If `concatenate_output` is True, a single tensor is returned. Else, a tuple with one tensor of dynamic
            inputs and one tensor of static inputs.
        """
        # transpose to [seq_length, batch_size, n_features]
        if self.embedding_type == 'full_model':
            data_type = 'x_d'
        elif self.embedding_type == 'forecast':
            data_type = 'x_f'
        elif self.embedding_type == 'hindcast':
            data_type = 'x_h'
        x_d = data[data_type].transpose(0, 1)

        if 'x_s' in data and 'x_one_hot' in data:
            x_s = torch.cat([data['x_s'], data['x_one_hot']], dim=-1)
        elif 'x_s' in data:
            x_s = data['x_s']
        elif 'x_one_hot' in data:
            x_s = data['x_one_hot']
        else:
            x_s = None

        # Don't run autoregressive inputs through the embedding layer. This does not work with NaN's
        if self._num_autoregression_inputs > 0:
            dynamics_out = self.dynamics_embedding(x_d[:, :, :-self._num_autoregression_inputs])
        else:
            dynamics_out = self.dynamics_embedding(x_d)

        statics_out = None
        if x_s is not None:
            statics_out = self.statics_embedding(x_s)

        if not concatenate_output:
            ret_val = dynamics_out, statics_out
        else:
            if statics_out is not None:
                statics_out = statics_out.unsqueeze(0).repeat(dynamics_out.shape[0], 1, 1)
                ret_val = torch.cat([dynamics_out, statics_out], dim=-1)
            else:
                ret_val = dynamics_out
            
            # Append autoregressive inputs to the end of the output.
            if self._num_autoregression_inputs:
                ret_val = torch.cat([ret_val, x_d[:, :, -self._num_autoregression_inputs:]], dim=-1)

        return ret_val

    def __getitem__(self, item: str) -> nn.Module:
        # required for dict-like access when freezing submodules' gradients in fine-tuning
        if item == "statics_embedding":
            return self.statics_embedding
        elif item == "dynamics_embedding":
            return self.dynamics_embedding
        else:
            raise KeyError(f"Cannot access {item} on InputLayer")



import logging
import pickle
import random
import re
import sys
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import torch
import xarray
from torch.utils.data import DataLoader
from tqdm import tqdm

from neuralhydrology.datasetzoo import get_dataset
from neuralhydrology.datasetzoo.basedataset import BaseDataset
from neuralhydrology.datautils.utils import get_frequency_factor, load_basin_file, load_scaler, sort_frequencies
from neuralhydrology.evaluation import plots
from neuralhydrology.evaluation.metrics import calculate_metrics, get_available_metrics
from neuralhydrology.evaluation.utils import load_basin_id_encoding, metrics_to_dataframe
from neuralhydrology.modelzoo import get_model
from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.training import get_loss_obj, get_regularization_obj
from neuralhydrology.training.logger import Logger
from neuralhydrology.utils.config import Config
from neuralhydrology.utils.errors import AllNaNError, NoEvaluationDataError

LOGGER = logging.getLogger(__name__)


class BaseTester(object):
    """Base class to run inference on a model.

    Use subclasses of this class to evaluate a trained model on its train, test, or validation period.
    For regression settings, `RegressionTester` is used; for uncertainty prediction, `UncertaintyTester`.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    run_dir : Path
        Path to the run directory.
    period : {'train', 'validation', 'test'}, optional
        The period to evaluate, by default 'test'.
    init_model : bool, optional
        If True, the model weights will be initialized with the checkpoint from the last available epoch in `run_dir`.
    """

    def __init__(self, cfg: Config, run_dir: Path, period: str = "test", init_model: bool = True):
        self.cfg = cfg
        self.run_dir = run_dir
        self.init_model = init_model
        if period in ["train", "validation", "test"]:
            self.period = period
        else:
            raise ValueError(f'Invalid period {period}. Must be one of ["train", "validation", "test"]')

        # determine device
        self._set_device()

        if self.init_model:
            self.model = get_model(cfg).to(self.device)

        self._disable_pbar = cfg.verbose == 0

        # pre-initialize variables, defined in class methods
        self.basins = None
        self.scaler = None
        self.id_to_int = {}
        self.additional_features = []

        # placeholder to store cached validation data
        self.cached_datasets = {}

        # initialize loss object to compute the loss of the evaluation data
        self.loss_obj = get_loss_obj(cfg)
        self.loss_obj.set_regularization_terms(get_regularization_obj(cfg=self.cfg))

        self._load_run_data()

    def _set_device(self):
        if self.cfg.device is not None:
            if "cuda" in self.cfg.device:
                gpu_id = int(self.cfg.device.split(':')[-1])
                if gpu_id > torch.cuda.device_count():
                    raise RuntimeError(f"This machine does not have GPU #{gpu_id} ")
                else:
                    self.device = torch.device(self.cfg.device)
            else:
                self.device = torch.device("cpu")
        else:
            self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    def _load_run_data(self):
        """Load run specific data from run directory"""

        # get list of basins
        self.basins = load_basin_file(getattr(self.cfg, f"{self.period}_basin_file"))

        # load feature scaler
        self.scaler = load_scaler(self.cfg.run_dir)

        # check for old scaler files, where the center/scale parameters had still old names
        if "xarray_means" in self.scaler.keys():
            self.scaler["xarray_feature_center"] = self.scaler.pop("xarray_means")
        if "xarray_stds" in self.scaler.keys():
            self.scaler["xarray_feature_scale"] = self.scaler.pop("xarray_stds")

        # load basin_id to integer dictionary for one-hot-encoding
        if self.cfg.use_basin_id_encoding:
            self.id_to_int = load_basin_id_encoding(self.cfg.run_dir)

        for file in self.cfg.additional_feature_files:
            with open(file, "rb") as fp:
                self.additional_features.append(pickle.load(fp))

    def _get_weight_file(self, epoch: int):
        """Get file path to weight file"""
        if epoch is None:
            weight_file = sorted(list(self.run_dir.glob('model_epoch*.pt')))[-1]
        else:
            weight_file = self.run_dir / f"model_epoch{str(epoch).zfill(3)}.pt"

        return weight_file

    def _load_weights(self, epoch: int = None):
        """Load weights of a certain (or the last) epoch into the model."""
        weight_file = self._get_weight_file(epoch)

        LOGGER.info(f"Using the model weights from {weight_file}")
        self.model.load_state_dict(torch.load(weight_file, map_location=self.device))

    def _get_dataset(self, basin: str) -> BaseDataset:
        """Get dataset for a single basin."""
        ds = get_dataset(cfg=self.cfg,
                         is_train=False,
                         period=self.period,
                         basin=basin,
                         additional_features=self.additional_features,
                         id_to_int=self.id_to_int,
                         scaler=self.scaler)
        return ds

    def evaluate(self,
                 epoch: int = None,
                 save_results: bool = True,
                 save_all_output: bool = False,
                 metrics: Union[list, dict] = [],
                 model: torch.nn.Module = None,
                 experiment_logger: Logger = None) -> dict:
        """Evaluate the model.
        
        Parameters
        ----------
        epoch : int, optional
            Define a specific epoch to evaluate. By default, the weights of the last epoch are used.
        save_results : bool, optional
            If True, stores the evaluation results in the run directory. By default, True.
        save_all_output : bool, optional
            If True, stores all of the model output in the run directory. By default, False.
        metrics : Union[list, dict], optional
            List of metrics to compute during evaluation. Can also be a dict that specifies per-target metrics
        model : torch.nn.Module, optional
            If a model is passed, this is used for validation.
        experiment_logger : Logger, optional
            Logger can be passed during training to log metrics

        Returns
        -------
        dict
            A dictionary containing one xarray per basin with the evaluation results.
        """
        if model is None:
            if self.init_model:
                self._load_weights(epoch=epoch)
                model = self.model
            else:
                raise RuntimeError("No model was initialized for the evaluation")

        # during validation, depending on settings, only evaluate on a random subset of basins
        basins = self.basins
        if self.period == "validation":
            if len(basins) > self.cfg.validate_n_random_basins:
                random.shuffle(basins)
                basins = basins[:self.cfg.validate_n_random_basins]

        # force model to train-mode when doing mc-dropout evaluation
        if self.cfg.mc_dropout:
            model.train()
        else:
            model.eval()

        results = defaultdict(dict)
        all_output = {basin: None for basin in basins}

        pbar = tqdm(basins, file=sys.stdout, disable=self._disable_pbar)
        pbar.set_description('# Validation' if self.period == "validation" else "# Evaluation")

        for basin in pbar:

            if self.cfg.cache_validation_data and basin in self.cached_datasets.keys():
                ds = self.cached_datasets[basin]
            else:
                try:
                    ds = self._get_dataset(basin)
                except NoEvaluationDataError as error:
                    # skip basin
                    continue
                if self.cfg.cache_validation_data and self.period == "validation":
                    self.cached_datasets[basin] = ds

            loader = DataLoader(ds, batch_size=self.cfg.batch_size, num_workers=0, collate_fn=ds.collate_fn)

            y_hat, y, dates, all_losses, all_output[basin] = self._evaluate(model, loader, ds.frequencies,
                                                                            save_all_output)

            # log loss of this basin plus number of samples in the logger to compute epoch aggregates later
            if experiment_logger is not None:
                experiment_logger.log_step(**{k: (v, len(loader)) for k, v in all_losses.items()})

            predict_last_n = self.cfg.predict_last_n
            seq_length = self.cfg.seq_length
            # if predict_last_n/seq_length are int, there's only one frequency
            if isinstance(predict_last_n, int):
                predict_last_n = {ds.frequencies[0]: predict_last_n}
            if isinstance(seq_length, int):
                seq_length = {ds.frequencies[0]: seq_length}
            lowest_freq = sort_frequencies(ds.frequencies)[0]
            for freq in ds.frequencies:
                if predict_last_n[freq] == 0:
                    continue  # this frequency is not being predicted
                results[basin][freq] = {}

                # rescale observations
                feature_scaler = self.scaler["xarray_feature_scale"][self.cfg.target_variables].to_array().values
                feature_center = self.scaler["xarray_feature_center"][self.cfg.target_variables].to_array().values
                y_freq = y[freq] * feature_scaler + feature_center
                # rescale predictions
                if y_hat[freq].ndim == 3 or (len(feature_scaler) == 1):
                    y_hat_freq = y_hat[freq] * feature_scaler + feature_center
                elif y_hat[freq].ndim == 4:
                    # if y_hat has 4 dim and we have multiple features we expand the dimensions for scaling
                    feature_scaler = np.expand_dims(feature_scaler, (0, 1, 3))
                    feature_center = np.expand_dims(feature_center, (0, 1, 3))
                    y_hat_freq = y_hat[freq] * feature_scaler + feature_center
                else:
                    raise RuntimeError(f"Simulations have {y_hat[freq].ndim} dimension. Only 3 and 4 are supported.")

                # Create data_vars dictionary for the xarray.Dataset
                data_vars = self._create_xarray_data_vars(y_hat_freq, y_freq)

                # freq_range are the steps of the current frequency at each lowest-frequency step
                frequency_factor = int(get_frequency_factor(lowest_freq, freq))

                # Create coords dictionary for the xarray.Dataset. 'date' can be directly infered from the dates
                # dictionary. We index the sample by the date of the last timestep of the sequence. The 'time_step'
                # index that specifies the position in the output sequence (relative to the end) can be inferred by
                # computing the timedelta of the dates. To account for predict_last_n > 1 and multi-freq stuff, we
                # need to add the frequency factor and remove 1 (to start at zero).
                coords = {
                    'date':
                        dates[lowest_freq][:, -1],
                    'time_step': ((dates[freq][0, :] - dates[freq][0, -1]) / pd.Timedelta(freq)).astype(np.int64) +
                                 frequency_factor - 1
                }
                xr = xarray.Dataset(data_vars=data_vars, coords=coords)
                xr = xr.reindex({
                    'date':
                        pd.DatetimeIndex(pd.date_range(xr["date"].values[0], xr["date"].values[-1], freq=lowest_freq),
                                         name='date')
                })
                results[basin][freq]['xr'] = xr

                # create datetime range at the current frequency
                freq_date_range = pd.date_range(start=dates[lowest_freq][0, -1], end=dates[freq][-1, -1], freq=freq)
                # remove datetime steps that are not being predicted from the datetime range
                mask = np.ones(frequency_factor).astype(bool)
                mask[:-predict_last_n[freq]] = False
                freq_date_range = freq_date_range[np.tile(mask, len(xr['date']))]

                # only warn once per freq
                if frequency_factor < predict_last_n[freq] and basin == basins[0]:
                    tqdm.write(f'Metrics for {freq} are calculated over last {frequency_factor} elements only. '
                               f'Ignoring {predict_last_n[freq] - frequency_factor} predictions per sequence.')

                if metrics:
                    for target_variable in self.cfg.target_variables:
                        # stack dates and time_steps so we don't just evaluate every 24H when use_frequencies=[1D, 1H]
                        obs = xr.isel(time_step=slice(-frequency_factor, None)) \
                            .stack(datetime=['date', 'time_step']) \
                            .drop_vars({'datetime', 'date', 'time_step'})[f"{target_variable}_obs"]
                        obs['datetime'] = freq_date_range
                        # check if there are observations for this period
                        if not all(obs.isnull()):
                            sim = xr.isel(time_step=slice(-frequency_factor, None)) \
                                .stack(datetime=['date', 'time_step']) \
                                .drop_vars({'datetime', 'date', 'time_step'})[f"{target_variable}_sim"]
                            sim['datetime'] = freq_date_range

                            # clip negative predictions to zero, if variable is listed in config 'clip_target_to_zero'
                            if target_variable in self.cfg.clip_targets_to_zero:
                                sim = xarray.where(sim < 0, 0, sim)

                            if 'samples' in sim.dims:
                                sim = sim.mean(dim='samples')

                            var_metrics = metrics if isinstance(metrics, list) else metrics[target_variable]
                            if 'all' in var_metrics:
                                var_metrics = get_available_metrics()
                            try:
                                values = calculate_metrics(obs, sim, metrics=var_metrics, resolution=freq)
                            except AllNaNError as err:
                                msg = f'Basin {basin} ' \
                                    + (f'{target_variable} ' if len(self.cfg.target_variables) > 1 else '') \
                                    + (f'{freq} ' if len(ds.frequencies) > 1 else '') \
                                    + str(err)
                                LOGGER.warning(msg)
                                values = {metric: np.nan for metric in var_metrics}

                            # add variable identifier to metrics if needed
                            if len(self.cfg.target_variables) > 1:
                                values = {f"{target_variable}_{key}": val for key, val in values.items()}
                            # add frequency identifier to metrics if needed
                            if len(ds.frequencies) > 1:
                                values = {f"{key}_{freq}": val for key, val in values.items()}
                            if experiment_logger is not None:
                                experiment_logger.log_step(**values)
                            for k, v in values.items():
                                results[basin][freq][k] = v

        # convert default dict back to normal Python dict to avoid unexpected behavior when trying to access
        # a non-existing basin
        results = dict(results)

        if (self.period == "validation") and (self.cfg.log_n_figures > 0) and (experiment_logger
                                                                               is not None) and results:
            self._create_and_log_figures(results, experiment_logger, epoch)

        # save model output to file, if requested
        results_to_save = None
        states_to_save = None
        if save_results:
            results_to_save = results
        if save_all_output:
            states_to_save = all_output
        if save_results or save_all_output:
            self._save_results(results=results_to_save, states=states_to_save, epoch=epoch)

        return results

    def _create_and_log_figures(self, results: dict, experiment_logger: Logger, epoch: int):
        basins = list(results.keys())
        random.shuffle(basins)
        for target_var in self.cfg.target_variables:
            max_figures = min(self.cfg.validate_n_random_basins, self.cfg.log_n_figures, len(basins))
            for freq in results[basins[0]].keys():
                figures = []
                for i in range(max_figures):
                    xr = results[basins[i]][freq]['xr']
                    obs = xr[f"{target_var}_obs"].values
                    sim = xr[f"{target_var}_sim"].values
                    # clip negative predictions to zero, if variable is listed in config 'clip_target_to_zero'
                    if target_var in self.cfg.clip_targets_to_zero:
                        sim = xarray.where(sim < 0, 0, sim)
                    figures.append(
                        self._get_plots(
                            obs, sim, title=f"{target_var} - Basin {basins[i]} - Epoch {epoch} - Frequency {freq}")[0])
                # make sure the preamble is a valid file name
                experiment_logger.log_figures(figures, freq, preamble=re.sub(r"[^A-Za-z0-9\._\-]+", "", target_var))

    def _save_results(self, results: Optional[dict], states: Optional[dict] = None, epoch: int = None):
        """Store results in various formats to disk.
        
        Developer note: We cannot store the time series data (the xarray objects) as netCDF file but have to use
        pickle as a wrapper. The reason is that netCDF files have special constraints on the characters/symbols that can
        be used as variable names. However, for convenience we will store metrics, if calculated, in a separate csv-file.
        """
        # use name of weight file as part of the result folder name
        weight_file = self._get_weight_file(epoch=epoch)

        # make sure the parent directory exists
        parent_directory = self.run_dir / self.period / weight_file.stem
        parent_directory.mkdir(parents=True, exist_ok=True)

        # save metrics any time this function is called, as long as they exist
        if self.cfg.metrics and results is not None:
            df = metrics_to_dataframe(results, self.cfg.metrics)
            metrics_file = parent_directory / f"{self.period}_metrics.csv"
            df.to_csv(metrics_file)
            LOGGER.info(f"Stored metrics at {metrics_file}")

        # store all results packed as pickle file
        if results is not None:
            result_file = parent_directory / f"{self.period}_results.p"
            with result_file.open("wb") as fp:
                pickle.dump(results, fp)
            LOGGER.info(f"Stored results at {result_file}")

        # store all model output packed as pickle file
        if states is not None:
            result_file = parent_directory / f"{self.period}_all_output.p"
            with result_file.open("wb") as fp:
                pickle.dump(states, fp)
            LOGGER.info(f"Stored states at {result_file}")

    def _evaluate(self, model: BaseModel, loader: DataLoader, frequencies: List[str], save_all_output: bool = False):
        """Evaluate model"""
        predict_last_n = self.cfg.predict_last_n
        if isinstance(predict_last_n, int):
            predict_last_n = {frequencies[0]: predict_last_n}  # if predict_last_n is int, there's only one frequency

        preds, obs, dates, all_output = {}, {}, {}, {}
        losses = []
        with torch.no_grad():
            for data in loader:

                for key in data:
                    if not key.startswith('date'):
                        data[key] = data[key].to(self.device)
                data = model.pre_model_hook(data, is_train=False)
                predictions, loss = self._get_predictions_and_loss(model, data)

                if all_output:
                    for key, value in predictions.items():
                        if value is not None and type(value) != dict:
                            all_output[key].append(value.detach().cpu().numpy())
                elif save_all_output:
                    all_output = {
                        key: [value.detach().cpu().numpy()]
                        for key, value in predictions.items()
                        if value is not None and type(value) != dict
                    }

                for freq in frequencies:
                    if predict_last_n[freq] == 0:
                        continue  # no predictions for this frequency
                    freq_key = '' if len(frequencies) == 1 else f'_{freq}'
                    y_hat_sub, y_sub = self._subset_targets(model, data, predictions, predict_last_n[freq], freq_key)
                    # Date subsetting is universal across all models and thus happens here.
                    date_sub = data[f'date{freq_key}'][:, -predict_last_n[freq]:]

                    if freq not in preds:
                        preds[freq] = y_hat_sub.detach().cpu()
                        obs[freq] = y_sub.cpu()
                        dates[freq] = date_sub
                    else:
                        preds[freq] = torch.cat((preds[freq], y_hat_sub.detach().cpu()), 0)
                        obs[freq] = torch.cat((obs[freq], y_sub.detach().cpu()), 0)
                        dates[freq] = np.concatenate((dates[freq], date_sub), axis=0)

                losses.append(loss)

            for freq in preds.keys():
                preds[freq] = preds[freq].numpy()
                obs[freq] = obs[freq].numpy()

        # concatenate all output variables (currently a dict-of-dicts) into a single-level dict
        for key, list_of_data in all_output.items():
            all_output[key] = np.concatenate(list_of_data, 0)

        # set to NaN explicitly if all losses are NaN to avoid RuntimeWarning
        mean_losses = {}
        if len(losses) == 0:
            mean_losses['loss'] = np.nan
        else:
            for loss_name in losses[0].keys():
                loss_values = [loss[loss_name] for loss in losses]
                mean_losses[loss_name] = np.nanmean(loss_values) if not np.all(np.isnan(loss_values)) else np.nan

        return preds, obs, dates, mean_losses, all_output

    def _get_predictions_and_loss(self, model: BaseModel, data: Dict[str, torch.Tensor]) -> Tuple[torch.Tensor, float]:
        predictions = model(data)
        _, all_losses = self.loss_obj(predictions, data)
        return predictions, {k: v.item() for k, v in all_losses.items()}

    def _subset_targets(self, model: BaseModel, data: Dict[str, torch.Tensor], predictions: np.ndarray,
                        predict_last_n: int, freq: str):
        raise NotImplementedError

    def _create_xarray_data_vars(self, y_hat: np.ndarray, y: np.ndarray):
        raise NotImplementedError

    def _get_plots(self, qobs: np.ndarray, qsim: np.ndarray, title: str):
        raise NotImplementedError


class RegressionTester(BaseTester):
    """Tester class to run inference on a regression model.

    Use the `evaluate` method of this class to evaluate a trained model on its train, test, or validation period.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    run_dir : Path
        Path to the run directory.
    period : {'train', 'validation', 'test'}
        The period to evaluate.
    init_model : bool, optional
        If True, the model weights will be initialized with the checkpoint from the last available epoch in `run_dir`.
    """

    def __init__(self, cfg: Config, run_dir: Path, period: str = "test", init_model: bool = True):
        super(RegressionTester, self).__init__(cfg, run_dir, period, init_model)

    def _subset_targets(self, model: BaseModel, data: Dict[str, torch.Tensor], predictions: np.ndarray,
                        predict_last_n: np.ndarray, freq: str):
        y_hat_sub = predictions[f'y_hat{freq}'][:, -predict_last_n:, :]
        y_sub = data[f'y{freq}'][:, -predict_last_n:, :]
        return y_hat_sub, y_sub

    def _create_xarray_data_vars(self, y_hat: np.ndarray, y: np.ndarray):
        data = {}
        for i, var in enumerate(self.cfg.target_variables):
            data[f"{var}_obs"] = (('date', 'time_step'), y[:, :, i])
            data[f"{var}_sim"] = (('date', 'time_step'), y_hat[:, :, i])
        return data

    def _get_plots(self, qobs: np.ndarray, qsim: np.ndarray, title: str):
        return plots.regression_plot(qobs, qsim, title)


class UncertaintyTester(BaseTester):
    """Tester class to run inference on an uncertainty model.

    Use the `evaluate` method of this class to evaluate a trained model on its train, test, or validation period.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    run_dir : Path
        Path to the run directory.
    period : {'train', 'validation', 'test'}
        The period to evaluate.
    init_model : bool, optional
        If True, the model weights will be initialized with the checkpoint from the last available epoch in `run_dir`.
    """

    def __init__(self, cfg: Config, run_dir: Path, period: str = "test", init_model: bool = True):
        super(UncertaintyTester, self).__init__(cfg, run_dir, period, init_model)

    def _get_predictions_and_loss(self, model: BaseModel, data: Dict[str, torch.Tensor]) -> Tuple[torch.Tensor, float]:
        outputs = model(data)
        _, all_losses = self.loss_obj(outputs, data)
        predictions = model.sample(data, self.cfg.n_samples)
        model.eval()
        return predictions, {k: v.item() for k, v in all_losses.items()}

    def _subset_targets(self,
                        model: BaseModel,
                        data: Dict[str, torch.Tensor],
                        predictions: np.ndarray,
                        predict_last_n: int,
                        freq: str = None):
        y_hat_sub = predictions[f'y_hat{freq}'][:, -predict_last_n:, :]
        y_sub = data[f'y{freq}'][:, -predict_last_n:, :]
        return y_hat_sub, y_sub

    def _create_xarray_data_vars(self, y_hat: np.ndarray, y: np.ndarray):
        data = {}
        for i, var in enumerate(self.cfg.target_variables):
            data[f"{var}_obs"] = (('date', 'time_step'), y[:, :, i])
            data[f"{var}_sim"] = (('date', 'time_step', 'samples'), y_hat[:, :, i, :])
        return data

    def _get_plots(self, qobs: np.ndarray, qsim: np.ndarray, title: str):
        return plots.uncertainty_plot(qobs, qsim, title)



from pathlib import Path
from typing import List, Dict, Union

import numpy as np
import pandas as pd
import xarray
from tqdm import tqdm

from neuralhydrology.datasetzoo.basedataset import BaseDataset
from neuralhydrology.utils.config import Config


class CamelsAUS(BaseDataset):
    """Data set class for the CAMELS-AUS dataset by [#]_.

    For more efficient data loading during model training/evaluating, this dataset class expects the CAMELS-AUS dataset
    in a processed format. Specifically, this dataset class works with per-basin csv files that contain all 
    timeseries data combined. Use the :func:`preprocess_camels_aus_dataset` function to process the original dataset 
    layout into this format.

    Parameters
    ----------
    cfg : Config
        The run configuration.
    is_train : bool 
        Defines if the dataset is used for training or evaluating. If True (training), means/stds for each feature
        are computed and stored to the run directory. If one-hot encoding is used, the mapping for the one-hot encoding 
        is created and also stored to disk. If False, a `scaler` input is expected and similarly the `id_to_int` input
        if one-hot encoding is used. 
    period : {'train', 'validation', 'test'}
        Defines the period for which the data will be loaded
    basin : str, optional
        If passed, the data for only this basin will be loaded. Otherwise the basin(s) are read from the appropriate
        basin file, corresponding to the `period`.
    additional_features : List[Dict[str, pd.DataFrame]], optional
        List of dictionaries, mapping from a basin id to a pandas DataFrame. This DataFrame will be added to the data
        loaded from the dataset, and all columns are available as 'dynamic_inputs', 'evolving_attributes' and
        'target_variables'
    id_to_int : Dict[str, int], optional
        If the config argument 'use_basin_id_encoding' is True in the config and period is either 'validation' or 
        'test', this input is required. It is a dictionary, mapping from basin id to an integer (the one-hot encoding).
    scaler : Dict[str, Union[pd.Series, xarray.DataArray]], optional
        If period is either 'validation' or 'test', this input is required. It contains the centering and scaling
        for each feature and is stored to the run directory during training (train_data/train_data_scaler.yml).

    References
    ----------
    .. [#] Fowler, K. J. A., Acharya, S. C., Addor, N., Chou, C., and Peel, M. C.: CAMELS-AUS: hydrometeorological time
        series and landscape attributes for 222 catchments in Australia, Earth Syst. Sci. Data, 13, 3847-3867, 
        https://doi.org/10.5194/essd-13-3847-2021, 2021. 
    """

    def __init__(self,
                 cfg: Config,
                 is_train: bool,
                 period: str,
                 basin: str = None,
                 additional_features: List[Dict[str, pd.DataFrame]] = [],
                 id_to_int: Dict[str, int] = {},
                 scaler: Dict[str, Union[pd.Series, xarray.DataArray]] = {}):
        super(CamelsAUS, self).__init__(cfg=cfg,
                                        is_train=is_train,
                                        period=period,
                                        basin=basin,
                                        additional_features=additional_features,
                                        id_to_int=id_to_int,
                                        scaler=scaler)

    def _load_basin_data(self, basin: str) -> pd.DataFrame:
        """Load input and output data from text files."""
        return load_camels_aus_timeseries(data_dir=self.cfg.data_dir, basin=basin)

    def _load_attributes(self) -> pd.DataFrame:
        """Load static catchment attributes."""
        return load_camels_aus_attributes(self.cfg.data_dir, basins=self.basins)


def load_camels_aus_timeseries(data_dir: Path, basin: str) -> pd.DataFrame:
    """Load the time series data for one basin of the CAMELS-AUS data set.

    Parameters
    ----------
    data_dir : Path
        Path to the CAMELS-AUS directory. This folder must contain a folder called 'preprocessed' containing the 
        per-basin csv files created by :func:`preprocess_camels_aus_dataset`.
    basin : str
        Basin identifier as string.

    Returns
    -------
    pd.DataFrame
        Time-indexed DataFrame, containing the time series data (forcings + discharge) data.
        
    Raises
    ------
    FileNotFoundError
        If no sub-folder called 'preprocessed' exists within the root directory of the CAMELS-AUS dataset.
    """
    preprocessed_dir = data_dir / "preprocessed"
    if not preprocessed_dir.is_dir():
        msg = [
            f"No preprocessed data directory found at {preprocessed_dir}. Use preprocessed_camels_aus_dataset in ",
            "neuralhydrology.datasetzoo.camelsaus to preprocess the CAMELS-AUS data set once into per-basin files."
        ]
        raise FileNotFoundError("".join(msg))
    basin_file = preprocessed_dir / f"{basin}.csv"
    df = pd.read_csv(basin_file, index_col='date', parse_dates=['date'])
    return df


def load_camels_aus_attributes(data_dir: Path, basins: List[str] = []) -> pd.DataFrame:
    """Load CAMELS-AUS attributes.

    Parameters
    ----------
    data_dir : Path
        Path to the CAMELS-AUS directory. Assumes that CAMELS_AUS_Attributes&Indices_MasterTable.csv is located in the
        data directory root folder.
    basins : List[str], optional
        If passed, return only attributes for the basins specified in this list. Otherwise, the attributes of all basins
        are returned.

    Returns
    -------
    pd.DataFrame
        Basin-indexed DataFrame, containing the attributes as columns.
    """
    attributes_file = data_dir / 'CAMELS_AUS_Attributes&Indices_MasterTable.csv'

    df = pd.read_csv(attributes_file, index_col="station_id")

    # convert all columns, where possible, to numeric
    df = df.apply(pd.to_numeric, errors='ignore')

    # convert the two columns specifying record period start and end to datetime format
    df["start_date"] = pd.to_datetime(df["start_date"], format="%Y%m%d")
    df["end_date"] = pd.to_datetime(df["end_date"], format="%Y%m%d")

    if basins:
        if any(b not in df.index for b in basins):
            raise ValueError('Some basins are missing static attributes.')
        df = df.loc[basins]

    return df


def preprocess_camels_aus_dataset(data_dir: Path):
    """Preprocess CAMELS-AUS data set and create per-basin files for more flexible and faster data loading.
    
    This function will read-in all time series text files and create per-basin csv files in a new subfolder called
    "preprocessed".
    
    Parameters
    ----------
    data_dir : Path
        Path to the CAMELS-AUS data set. Expects different subfolders with the original names, specifically
        '05_hydrometeorology' and '03_streamflow'.

    Raises
    ------
    FileExistsError
        If a sub-folder called 'preprocessed' already exists in `data_dir`.
    """
    # check if data has already been pre-processed other-wise create dst folder
    dst_dir = data_dir / "preprocessed"
    if dst_dir.is_dir():
        raise FileExistsError(
            "Subdirectory 'preprocessed' already exists. Delete this folder if you want to reprocess the data.")
    dst_dir.mkdir()

    # Load all different forcing files into memory
    forcing_dir = data_dir / "05_hydrometeorology"
    files = [f for f in forcing_dir.glob('**/*.csv') if f.name != 'ClimaticIndices.csv']
    dfs = {}
    for f in tqdm(files, desc="Read meteorological forcing data into memory"):
        df = pd.read_csv(f)
        df["date"] = pd.to_datetime(df.year.map(str) + "/" + df.month.map(str) + "/" + df.day.map(str),
                                    format="%Y/%m/%d")
        df = df.set_index('date')
        dfs[f.stem] = df

    # Load streamflow data into memory and replace invalid measurements (-99) with NaNs
    print("Read streamflow data into memory.")
    df = pd.read_csv(data_dir / "03_streamflow" / "streamflow_mmd.csv")
    df["date"] = pd.to_datetime(df.year.map(str) + "/" + df.month.map(str) + "/" + df.day.map(str), format="%Y/%m/%d")
    df = df.set_index('date')
    df[df < 0] = np.nan
    dfs["streamflow_mmd"] = df

    # Extract list of basins from column names
    basins = [c for c in dfs["streamflow_mmd"].columns if c not in ['year', 'month', 'day']]

    # Create per-basin dataframes by combining all forcing variables + streamflow and save to disk.
    for basin in tqdm(basins, desc="Create per-basin dataframes and save data to disk."):
        data = {}
        for key, df in dfs.items():
            data[key] = df[basin]
        df = pd.DataFrame(data)
        df.to_csv(dst_dir / f"{basin}.csv")

    print(f"Finished processing the CAMELS-AUS data set. Resulting per-basin csv files have been stored at {dst_dir}")



import logging
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
from scipy import stats, signal
from xarray.core.dataarray import DataArray

from neuralhydrology.datautils import utils
from neuralhydrology.utils.errors import AllNaNError

LOGGER = logging.getLogger(__name__)


def get_available_metrics() -> List[str]:
    """Get list of available metrics.

    Returns
    -------
    List[str]
        List of implemented metric names.
    """
    metrics = [
        "NSE", "MSE", "RMSE", "KGE", "Alpha-NSE", "Pearson-r", "Beta-KGE", "Beta-NSE", "FHV", "FMS", "FLV",
        "Peak-Timing", "Missed-Peaks", "Peak-MAPE"
    ]
    return metrics


def _validate_inputs(obs: DataArray, sim: DataArray):
    if obs.shape != sim.shape:
        raise RuntimeError("Shapes of observations and simulations must match")

    if (len(obs.shape) > 1) and (obs.shape[1] > 1):
        raise RuntimeError("Metrics only defined for time series (1d or 2d with second dimension 1)")


def _mask_valid(obs: DataArray, sim: DataArray) -> Tuple[DataArray, DataArray]:
    # mask of invalid entries. NaNs in simulations can happen during validation/testing
    idx = (~sim.isnull()) & (~obs.isnull())

    obs = obs[idx]
    sim = sim[idx]

    return obs, sim


def _get_fdc(da: DataArray) -> np.ndarray:
    return da.sortby(da, ascending=False).values


def nse(obs: DataArray, sim: DataArray) -> float:
    r"""Calculate Nash-Sutcliffe Efficiency [#]_
    
    Nash-Sutcliffe Efficiency is the R-square between observed and simulated discharge.
    
    .. math:: \text{NSE} = 1 - \frac{\sum_{t=1}^{T}(Q_m^t - Q_o^t)^2}{\sum_{t=1}^T(Q_o^t - \overline{Q}_o)^2},
    
    where :math:`Q_m` are the simulations (here, `sim`) and :math:`Q_o` are observations (here, `obs`).
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.

    Returns
    -------
    float
        Nash-Sutcliffe Efficiency 
        
    References
    ----------
    .. [#] Nash, J. E.; Sutcliffe, J. V. (1970). "River flow forecasting through conceptual models part I - A 
        discussion of principles". Journal of Hydrology. 10 (3): 282-290. doi:10.1016/0022-1694(70)90255-6.

    """

    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    denominator = ((obs - obs.mean())**2).sum()
    numerator = ((sim - obs)**2).sum()

    value = 1 - numerator / denominator

    return float(value)


def mse(obs: DataArray, sim: DataArray) -> float:
    r"""Calculate mean squared error.
    
    .. math:: \text{MSE} = \frac{1}{T}\sum_{t=1}^T (\widehat{y}_t - y_t)^2,
    
    where :math:`\widehat{y}` are the simulations (here, `sim`) and :math:`y` are observations 
    (here, `obs`).
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.

    Returns
    -------
    float
        Mean squared error. 

    """

    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    return float(((sim - obs)**2).mean())


def rmse(obs: DataArray, sim: DataArray) -> float:
    r"""Calculate root mean squared error.
    
    .. math:: \text{RMSE} = \sqrt{\frac{1}{T}\sum_{t=1}^T (\widehat{y}_t - y_t)^2},
    
    where :math:`\widehat{y}` are the simulations (here, `sim`) and :math:`y` are observations 
    (here, `obs`).
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.

    Returns
    -------
    float
        Root mean sqaured error.

    """

    return np.sqrt(mse(obs, sim))


def alpha_nse(obs: DataArray, sim: DataArray) -> float:
    r"""Calculate the alpha NSE decomposition [#]_
    
    The alpha NSE decomposition is the fraction of the standard deviations of simulations and observations.
    
    .. math:: \alpha = \frac{\sigma_s}{\sigma_o},
    
    where :math:`\sigma_s` is the standard deviation of the simulations (here, `sim`) and :math:`\sigma_o` is the 
    standard deviation of the observations (here, `obs`).
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.

    Returns
    -------
    float
        Alpha NSE decomposition.
        
    References
    ----------
    .. [#] Gupta, H. V., Kling, H., Yilmaz, K. K., & Martinez, G. F. (2009). Decomposition of the mean squared error 
        and NSE performance criteria: Implications for improving hydrological modelling. Journal of hydrology, 377(1-2),
        80-91.

    """

    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    return float(sim.std() / obs.std())


def beta_nse(obs: DataArray, sim: DataArray) -> float:
    r"""Calculate the beta NSE decomposition [#]_

    The beta NSE decomposition is the difference of the mean simulation and mean observation divided by the standard 
    deviation of the observations.

    .. math:: \beta = \frac{\mu_s - \mu_o}{\sigma_o},
    
    where :math:`\mu_s` is the mean of the simulations (here, `sim`), :math:`\mu_o` is the mean of the observations 
    (here, `obs`) and :math:`\sigma_o` the standard deviation of the observations.

    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.

    Returns
    -------
    float
        Beta NSE decomposition.

    References
    ----------
    .. [#] Gupta, H. V., Kling, H., Yilmaz, K. K., & Martinez, G. F. (2009). Decomposition of the mean squared error 
        and NSE performance criteria: Implications for improving hydrological modelling. Journal of hydrology, 377(1-2),
        80-91.

    """
    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    return float((sim.mean() - obs.mean()) / obs.std())


def beta_kge(obs: DataArray, sim: DataArray) -> float:
    r"""Calculate the beta KGE term [#]_
    
    The beta term of the Kling-Gupta Efficiency is defined as the fraction of the means.
    
    .. math:: \beta_{\text{KGE}} = \frac{\mu_s}{\mu_o},
    
    where :math:`\mu_s` is the mean of the simulations (here, `sim`) and :math:`\mu_o` is the mean of the observations 
    (here, `obs`).
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.

    Returns
    -------
    float
        Beta NSE decomposition.

    References
    ----------
    .. [#] Gupta, H. V., Kling, H., Yilmaz, K. K., & Martinez, G. F. (2009). Decomposition of the mean squared error 
        and NSE performance criteria: Implications for improving hydrological modelling. Journal of hydrology, 377(1-2),
        80-91.

    """
    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    return float(sim.mean() / obs.mean())


def kge(obs: DataArray, sim: DataArray, weights: List[float] = [1., 1., 1.]) -> float:
    r"""Calculate the Kling-Gupta Efficieny [#]_
    
    .. math:: 
        \text{KGE} = 1 - \sqrt{[ s_r (r - 1)]^2 + [s_\alpha ( \alpha - 1)]^2 + 
            [s_\beta(\beta_{\text{KGE}} - 1)]^2},
            
    where :math:`r` is the correlation coefficient, :math:`\alpha` the :math:`\alpha`-NSE decomposition, 
    :math:`\beta_{\text{KGE}}` the fraction of the means and :math:`s_r, s_\alpha, s_\beta` the corresponding weights
    (here the three float values in the `weights` parameter).
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.
    weights : List[float]
        Weighting factors of the 3 KGE parts, by default each part has a weight of 1.

    Returns
    -------
    float
        Kling-Gupta Efficiency
    
    References
    ----------
    .. [#] Gupta, H. V., Kling, H., Yilmaz, K. K., & Martinez, G. F. (2009). Decomposition of the mean squared error 
        and NSE performance criteria: Implications for improving hydrological modelling. Journal of hydrology, 377(1-2),
        80-91.

    """
    if len(weights) != 3:
        raise ValueError("Weights of the KGE must be a list of three values")

    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    if len(obs) < 2:
        return np.nan

    r, _ = stats.pearsonr(obs.values, sim.values)

    alpha = sim.std() / obs.std()
    beta = sim.mean() / obs.mean()

    value = (weights[0] * (r - 1)**2 + weights[1] * (alpha - 1)**2 + weights[2] * (beta - 1)**2)

    return 1 - np.sqrt(float(value))


def pearsonr(obs: DataArray, sim: DataArray) -> float:
    """Calculate pearson correlation coefficient (using scipy.stats.pearsonr)

    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.

    Returns
    -------
    float
        Pearson correlation coefficient

    """

    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    if len(obs) < 2:
        return np.nan

    r, _ = stats.pearsonr(obs.values, sim.values)

    return float(r)


def fdc_fms(obs: DataArray, sim: DataArray, lower: float = 0.2, upper: float = 0.7) -> float:
    r"""Calculate the slope of the middle section of the flow duration curve [#]_
    
    .. math:: 
        \%\text{BiasFMS} = \frac{\left | \log(Q_{s,\text{lower}}) - \log(Q_{s,\text{upper}}) \right | - 
            \left | \log(Q_{o,\text{lower}}) - \log(Q_{o,\text{upper}}) \right |}{\left | 
            \log(Q_{s,\text{lower}}) - \log(Q_{s,\text{upper}}) \right |} \times 100,
            
    where :math:`Q_{s,\text{lower/upper}}` corresponds to the FDC of the simulations (here, `sim`) at the `lower` and
    `upper` bound of the middle section and :math:`Q_{o,\text{lower/upper}}` similarly for the observations (here, 
    `obs`).
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.
    lower : float, optional
        Lower bound of the middle section in range ]0,1[, by default 0.2
    upper : float, optional
        Upper bound of the middle section in range ]0,1[, by default 0.7
        
    Returns
    -------
    float
        Slope of the middle section of the flow duration curve.
    
    References
    ----------
    .. [#] Yilmaz, K. K., Gupta, H. V., and Wagener, T. ( 2008), A process-based diagnostic approach to model 
        evaluation: Application to the NWS distributed hydrologic model, Water Resour. Res., 44, W09417, 
        doi:10.1029/2007WR006716. 
    """
    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    if len(obs) < 1:
        return np.nan

    if any([(x <= 0) or (x >= 1) for x in [upper, lower]]):
        raise ValueError("upper and lower have to be in range ]0,1[")

    if lower >= upper:
        raise ValueError("The lower threshold has to be smaller than the upper.")

    # get arrays of sorted (descending) discharges
    obs = _get_fdc(obs)
    sim = _get_fdc(sim)

    # for numerical reasons change 0s to 1e-6. Simulations can still contain negatives, so also reset those.
    sim[sim <= 0] = 1e-6
    obs[obs == 0] = 1e-6

    # calculate fms part by part
    qsm_lower = np.log(sim[np.round(lower * len(sim)).astype(int)])
    qsm_upper = np.log(sim[np.round(upper * len(sim)).astype(int)])
    qom_lower = np.log(obs[np.round(lower * len(obs)).astype(int)])
    qom_upper = np.log(obs[np.round(upper * len(obs)).astype(int)])

    fms = ((qsm_lower - qsm_upper) - (qom_lower - qom_upper)) / (qom_lower - qom_upper + 1e-6)

    return fms * 100


def fdc_fhv(obs: DataArray, sim: DataArray, h: float = 0.02) -> float:
    r"""Calculate the peak flow bias of the flow duration curve [#]_
    
    .. math:: \%\text{BiasFHV} = \frac{\sum_{h=1}^{H}(Q_{s,h} - Q_{o,h})}{\sum_{h=1}^{H}Q_{o,h}} \times 100,
    
    where :math:`Q_s` are the simulations (here, `sim`), :math:`Q_o` the observations (here, `obs`) and `H` is the upper
    fraction of flows of the FDC (here, `h`). 
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.
    h : float, optional
        Fraction of upper flows to consider as peak flows of range ]0,1[, be default 0.02.
        
    Returns
    -------
    float
        Peak flow bias.
    
    References
    ----------
    .. [#] Yilmaz, K. K., Gupta, H. V., and Wagener, T. ( 2008), A process-based diagnostic approach to model 
        evaluation: Application to the NWS distributed hydrologic model, Water Resour. Res., 44, W09417, 
        doi:10.1029/2007WR006716. 
    """
    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    if len(obs) < 1:
        return np.nan

    if (h <= 0) or (h >= 1):
        raise ValueError("h has to be in range ]0,1[. Consider small values, e.g. 0.02 for 2% peak flows")

    # get arrays of sorted (descending) discharges
    obs = _get_fdc(obs)
    sim = _get_fdc(sim)

    # subset data to only top h flow values
    obs = obs[:np.round(h * len(obs)).astype(int)]
    sim = sim[:np.round(h * len(sim)).astype(int)]

    fhv = np.sum(sim - obs) / np.sum(obs)

    return fhv * 100


def fdc_flv(obs: DataArray, sim: DataArray, l: float = 0.3) -> float:
    r"""Calculate the low flow bias of the flow duration curve [#]_
    
    .. math:: 
        \%\text{BiasFMS} = -1 \frac{\sum_{l=1}^{L}[\log(Q_{s,l}) - \log(Q_{s,L})] - \sum_{l=1}^{L}[\log(Q_{o,l})
            - \log(Q_{o,L})]}{\sum_{l=1}^{L}[\log(Q_{o,l}) - \log(Q_{o,L})]} \times 100,
    
    where :math:`Q_s` are the simulations (here, `sim`), :math:`Q_o` the observations (here, `obs`) and `L` is the lower
    fraction of flows of the FDC (here, `l`). 
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.
    l : float, optional
        Fraction of lower flows to consider as low flows of range ]0,1[, be default 0.3.
        
    Returns
    -------
    float
        Low flow bias.
    
    References
    ----------
    .. [#] Yilmaz, K. K., Gupta, H. V., and Wagener, T. ( 2008), A process-based diagnostic approach to model 
        evaluation: Application to the NWS distributed hydrologic model, Water Resour. Res., 44, W09417, 
        doi:10.1029/2007WR006716. 
    """
    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    if len(obs) < 1:
        return np.nan

    if (l <= 0) or (l >= 1):
        raise ValueError("l has to be in range ]0,1[. Consider small values, e.g. 0.3 for 30% low flows")

    # get arrays of sorted (descending) discharges
    obs = _get_fdc(obs)
    sim = _get_fdc(sim)

    # for numerical reasons change 0s to 1e-6. Simulations can still contain negatives, so also reset those.
    sim[sim <= 0] = 1e-6
    obs[obs == 0] = 1e-6

    obs = obs[-np.round(l * len(obs)).astype(int):]
    sim = sim[-np.round(l * len(sim)).astype(int):]

    # transform values to log scale
    obs = np.log(obs)
    sim = np.log(sim)

    # calculate flv part by part
    qsl = np.sum(sim - sim.min())
    qol = np.sum(obs - obs.min())

    flv = -1 * (qsl - qol) / (qol + 1e-6)

    return flv * 100


def mean_peak_timing(obs: DataArray,
                     sim: DataArray,
                     window: int = None,
                     resolution: str = '1D',
                     datetime_coord: str = None) -> float:
    """Mean difference in peak flow timing.
    
    Uses scipy.find_peaks to find peaks in the observed time series. Starting with all observed peaks, those with a
    prominence of less than the standard deviation of the observed time series are discarded. Next, the lowest peaks
    are subsequently discarded until all remaining peaks have a distance of at least 100 steps. Finally, the
    corresponding peaks in the simulated time series are searched in a window of size `window` on either side of the
    observed peaks and the absolute time differences between observed and simulated peaks is calculated.
    The final metric is the mean absolute time difference across all peaks. For more details, see Appendix of [#]_
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.
    window : int, optional
        Size of window to consider on each side of the observed peak for finding the simulated peak. That is, the total
        window length to find the peak in the simulations is :math:`2 * \\text{window} + 1` centered at the observed
        peak. The default depends on the temporal resolution, e.g. for a resolution of '1D', a window of 3 is used and 
        for a resolution of '1H' the the window size is 12.
    resolution : str, optional
        Temporal resolution of the time series in pandas format, e.g. '1D' for daily and '1H' for hourly.
    datetime_coord : str, optional
        Name of datetime coordinate. Tried to infer automatically if not specified.
        

    Returns
    -------
    float
        Mean peak time difference.

    References
    ----------
    .. [#] Kratzert, F., Klotz, D., Hochreiter, S., and Nearing, G. S.: A note on leveraging synergy in multiple 
        meteorological datasets with deep learning for rainfall-runoff modeling, Hydrol. Earth Syst. Sci. Discuss., 
        https://doi.org/10.5194/hess-2020-221, in review, 2020. 
    """
    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations (scipy's find_peaks doesn't guarantee correctness with NaNs)
    obs, sim = _mask_valid(obs, sim)

    # heuristic to get indices of peaks and their corresponding height.
    peaks, _ = signal.find_peaks(obs.values, distance=100, prominence=np.std(obs.values))

    # infer name of datetime index
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(obs)

    if window is None:
        # infer a reasonable window size
        window = max(int(utils.get_frequency_factor('12H', resolution)), 3)

    # evaluate timing
    timing_errors = []
    for idx in peaks:
        # skip peaks at the start and end of the sequence and peaks around missing observations
        # (NaNs that were removed in obs & sim would result in windows that span too much time).
        if (idx - window < 0) or (idx + window >= len(obs)) or (pd.date_range(obs[idx - window][datetime_coord].values,
                                                                              obs[idx + window][datetime_coord].values,
                                                                              freq=resolution).size != 2 * window + 1):
            continue

        # check if the value at idx is a peak (both neighbors must be smaller)
        if (sim[idx] > sim[idx - 1]) and (sim[idx] > sim[idx + 1]):
            peak_sim = sim[idx]
        else:
            # define peak around idx as the max value inside of the window
            values = sim[idx - window:idx + window + 1]
            peak_sim = values[values.argmax()]

        # get xarray object of qobs peak, for getting the date and calculating the datetime offset
        peak_obs = obs[idx]

        # calculate the time difference between the peaks
        delta = peak_obs.coords[datetime_coord] - peak_sim.coords[datetime_coord]

        timing_error = np.abs(delta.values / pd.to_timedelta(resolution))

        timing_errors.append(timing_error)

    return np.mean(timing_errors) if len(timing_errors) > 0 else np.nan


def missed_peaks(obs: DataArray,
                 sim: DataArray,
                 window: int = None,
                 resolution: str = '1D',
                 percentile: float = 80,
                 datetime_coord: str = None) -> float:
    """Fraction of missed peaks.
    
    Uses scipy.find_peaks to find peaks in the observed and simulated time series above a certain percentile. Counts
    the number of peaks in obs that do not exist in sim within the specified window.

    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.
    window : int, optional
        Size of window to consider on each side of the observed peak for finding the simulated peak. That is, the total
        window length to find the peak in the simulations is :math:`2 * \\text{window} + 1` centered at the observed
        peak. The default depends on the temporal resolution, e.g. for a resolution of '1D', a window of 1 is used and 
        for a resolution of '1H' the the window size is 12. Note that this is a different default window size than is
        used in the peak-timing metric for '1D'.
    resolution : str, optional
        Temporal resolution of the time series in pandas format, e.g. '1D' for daily and '1H' for hourly.
    percentile: float, optional
        Only consider peaks above this flow percentile (0, 100).
    datetime_coord : str, optional
        Name of datetime coordinate. Tried to infer automatically if not specified.

    Returns
    -------
    float
        Fraction of missed peaks.   
    """
    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations (scipy's find_peaks doesn't guarantee correctness with NaNs)
    obs, sim = _mask_valid(obs, sim)

    # minimum height of a peak, as defined by percentile, which can be passed
    min_obs_height = np.percentile(obs.values, percentile)
    min_sim_height = np.percentile(sim.values, percentile)

    # get time indices of peaks in obs and sim.
    peaks_obs_times, _ = signal.find_peaks(obs, distance=30, height=min_obs_height)
    peaks_sim_times, _ = signal.find_peaks(sim, distance=30, height=min_sim_height)

    if len(peaks_obs_times) == 0:
        return 0.

    # infer name of datetime index
    if datetime_coord is None:
        datetime_coord = utils.infer_datetime_coord(obs)

    # infer a reasonable window size
    if window is None:
        window = max(int(utils.get_frequency_factor('12H', resolution)), 1)

    # count missed peaks
    missed_events = 0

    for idx in peaks_obs_times:

        # skip peaks at the start and end of the sequence and peaks around missing observations
        # (NaNs that were removed in obs & sim would result in windows that span too much time).
        if (idx - window < 0) or (idx + window >= len(obs)) or (pd.date_range(obs[idx - window][datetime_coord].values,
                                                                              obs[idx + window][datetime_coord].values,
                                                                              freq=resolution).size != 2 * window + 1):
            continue

        nearby_peak_sim_index = np.where(np.abs(peaks_sim_times - idx) <= window)[0]
        if len(nearby_peak_sim_index) == 0:
            missed_events += 1

    return missed_events / len(peaks_obs_times)


def mean_absolute_percentage_peak_error(obs: DataArray, sim: DataArray) -> float:
    r"""Calculate the mean absolute percentage error (MAPE) for peaks

    .. math:: \text{MAPE}_\text{peak} = \frac{1}{P}\sum_{p=1}^{P} \left |\frac{Q_{s,p} - Q_{o,p}}{Q_{o,p}} \right | \times 100,

    where :math:`Q_{s,p}` are the simulated peaks (here, `sim`), :math:`Q_{o,p}` the observed peaks (here, `obs`) and
    `P` is the number of peaks.

    Uses scipy.find_peaks to find peaks in the observed time series. The observed peaks indices are used to subset
    observed and simulated flows. Finally, the MAPE metric is calculated as the mean absolute percentage error
    of observed peak flows and corresponding simulated flows.

    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.

    Returns
    -------
    float
        Mean absolute percentage error (MAPE) for peaks.
    """
    # verify inputs
    _validate_inputs(obs, sim)

    # get time series with only valid observations
    obs, sim = _mask_valid(obs, sim)

    # return np.nan if there are no valid observed or simulated values
    if obs.size == 0 or sim.size == 0:
        return np.nan

    # heuristic to get indices of peaks and their corresponding height.
    peaks, _ = signal.find_peaks(obs.values, distance=100, prominence=np.std(obs.values))

    # check if any peaks exist, otherwise return np.nan
    if peaks.size == 0:
        return np.nan

    # subset data to only peak values
    obs = obs[peaks].values
    sim = sim[peaks].values

    # calculate the mean absolute percentage peak error
    peak_mape = np.sum(np.abs((sim - obs) / obs)) / peaks.size * 100

    return peak_mape


def calculate_all_metrics(obs: DataArray,
                          sim: DataArray,
                          resolution: str = "1D",
                          datetime_coord: str = None) -> Dict[str, float]:
    """Calculate all metrics with default values.
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.
    resolution : str, optional
        Temporal resolution of the time series in pandas format, e.g. '1D' for daily and '1H' for hourly.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.
        
    Returns
    -------
    Dict[str, float]
        Dictionary with keys corresponding to metric name and values corresponding to metric values.

    Raises
    ------
    AllNaNError
        If all observations or all simulations are NaN.
    """
    _check_all_nan(obs, sim)

    results = {
        "NSE": nse(obs, sim),
        "MSE": mse(obs, sim),
        "RMSE": rmse(obs, sim),
        "KGE": kge(obs, sim),
        "Alpha-NSE": alpha_nse(obs, sim),
        "Beta-KGE": beta_kge(obs, sim),
        "Beta-NSE": beta_nse(obs, sim),
        "Pearson-r": pearsonr(obs, sim),
        "FHV": fdc_fhv(obs, sim),
        "FMS": fdc_fms(obs, sim),
        "FLV": fdc_flv(obs, sim),
        "Peak-Timing": mean_peak_timing(obs, sim, resolution=resolution, datetime_coord=datetime_coord),
        "Peak-MAPE": mean_absolute_percentage_peak_error(obs, sim)
    }

    return results


def calculate_metrics(obs: DataArray,
                      sim: DataArray,
                      metrics: List[str],
                      resolution: str = "1D",
                      datetime_coord: str = None) -> Dict[str, float]:
    """Calculate specific metrics with default values.
    
    Parameters
    ----------
    obs : DataArray
        Observed time series.
    sim : DataArray
        Simulated time series.
    metrics : List[str]
        List of metric names.
    resolution : str, optional
        Temporal resolution of the time series in pandas format, e.g. '1D' for daily and '1H' for hourly.
    datetime_coord : str, optional
        Datetime coordinate in the passed DataArray. Tried to infer automatically if not specified.

    Returns
    -------
    Dict[str, float]
        Dictionary with keys corresponding to metric name and values corresponding to metric values.

    Raises
    ------
    AllNaNError
        If all observations or all simulations are NaN.
    """
    if 'all' in metrics:
        return calculate_all_metrics(obs, sim, resolution=resolution)

    _check_all_nan(obs, sim)

    values = {}
    for metric in metrics:
        if metric.lower() == "nse":
            values["NSE"] = nse(obs, sim)
        elif metric.lower() == "mse":
            values["MSE"] = mse(obs, sim)
        elif metric.lower() == "rmse":
            values["RMSE"] = rmse(obs, sim)
        elif metric.lower() == "kge":
            values["KGE"] = kge(obs, sim)
        elif metric.lower() == "alpha-nse":
            values["Alpha-NSE"] = alpha_nse(obs, sim)
        elif metric.lower() == "beta-kge":
            values["Beta-KGE"] = beta_kge(obs, sim)
        elif metric.lower() == "beta-nse":
            values["Beta-NSE"] = beta_nse(obs, sim)
        elif metric.lower() == "pearson-r":
            values["Pearson-r"] = pearsonr(obs, sim)
        elif metric.lower() == "fhv":
            values["FHV"] = fdc_fhv(obs, sim)
        elif metric.lower() == "fms":
            values["FMS"] = fdc_fms(obs, sim)
        elif metric.lower() == "flv":
            values["FLV"] = fdc_flv(obs, sim)
        elif metric.lower() == "peak-timing":
            values["Peak-Timing"] = mean_peak_timing(obs, sim, resolution=resolution, datetime_coord=datetime_coord)
        elif metric.lower() == "missed-peaks":
            values["Missed-Peaks"] = missed_peaks(obs, sim, resolution=resolution, datetime_coord=datetime_coord)
        elif metric.lower() == "peak-mape":
            values["Peak-MAPE"] = mean_absolute_percentage_peak_error(obs, sim)
        else:
            raise RuntimeError(f"Unknown metric {metric}")

    return values


def _check_all_nan(obs: DataArray, sim: DataArray):
    """Check if all observations or simulations are NaN and raise an exception if this is the case.

    Raises
    ------
    AllNaNError
        If all observations or all simulations are NaN.
    """
    if all(obs.isnull()):
        raise AllNaNError("All observed values are NaN, thus metrics will be NaN, too.")
    if all(sim.isnull()):
        raise AllNaNError("All simulated values are NaN, thus metrics will be NaN, too.")



from typing import Dict, Union, Tuple
import torch
import torch.nn as nn
from neuralhydrology.utils.config import Config


class BaseConceptualModel(nn.Module):
    """Abstract base model class, don't use this class for model training.

    The purpose is to have some common operations that all conceptual models will need. Use subclasses of this class
    for training/evaluating different conceptual models, e.g. 'SHM'.

    Parameters
    ----------
    cfg : Config
        The run configuration.

    """

    def __init__(self, cfg: Config):
        super(BaseConceptualModel, self).__init__()
        self.cfg = cfg
        # Check if the dynamic_conceptual_inputs and the target_variables are in the custom normalization. This is
        # necessary as conceptual models are mass conservative.
        if any(item not in cfg.custom_normalization for item in cfg.dynamic_conceptual_inputs + cfg.target_variables):
            raise RuntimeError("dynamic_conceptual_inputs and target_variables require custom_normalization")

    def forward(self, x_conceptual: torch.Tensor, lstm_out: torch.Tensor) -> Dict[str, Union[torch.Tensor, Dict[str, torch.Tensor]]]:
        raise NotImplementedError

    def _get_dynamic_parameters_conceptual(self, lstm_out: torch.Tensor) -> Dict[str, torch.Tensor]:
        """Map the output of the data-driven part of the predefined ranges of the conceptual model that is being used.

        Parameters
        ----------
        lstm_out: torch.Tensor
            Tensor of size [batch_size, time_steps, n_param] that will be mapped to the predefined ranges of the
            conceptual model parameters to act as the dynamic parameterization.

        Returns
        -------
        Dict[str, torch.Tensor]
            Dynamic parameterization of the conceptual model.
        """
        dynamic_parameters = {}
        for index, (parameter_name, parameter_range) in enumerate(self.parameter_ranges.items()):
            range_t = torch.tensor(parameter_range, dtype=torch.float32, device=lstm_out.device)
            range_t = range_t.repeat(lstm_out.shape[0], 1)  # To run all the elements of the batch in parallel
            dynamic_parameters[parameter_name] = range_t[:, :1] + torch.sigmoid(lstm_out[:, :, index]) * (range_t[:, 1:] - range_t[:, :1])

        return dynamic_parameters

    def _initialize_information(self, conceptual_inputs: torch.Tensor) -> Tuple[Dict[str, torch.Tensor], torch.Tensor]:
        """Initialize the structures to store the time evolution of the internal states and the outflow of the conceptual
        model

        Parameters
        ----------
        conceptual_inputs: torch.Tensor
            Inputs of the conceptual model (dynamic forcings)

        Returns
        -------
        Tuple[Dict[str, torch.Tensor], torch.Tensor]
            - states: Dict[str, torch.Tensor]
                Dictionary to store the time evolution of the internal states (buckets) of the conceptual model
            - q_out: torch.Tensor
                Tensor to store the outputs of the conceptual model
        """

        states = {}
        # initialize dictionary to store the evolution of the states
        for name, value in self.initial_states.items():
            states[name] = torch.zeros((conceptual_inputs.shape[0], conceptual_inputs.shape[1]), dtype=torch.float32,
                                       device=conceptual_inputs.device)

        # initialize vectors to store the evolution of the outputs
        out = torch.zeros((conceptual_inputs.shape[0], conceptual_inputs.shape[1], len(self.cfg.target_variables)),
                          dtype=torch.float32, device=conceptual_inputs.device)

        return states, out


    @property
    def initial_states(self):
        raise NotImplementedError

    @property
    def parameter_ranges(self):
        raise NotImplementedError




import logging
from typing import Dict, Tuple, Union

import numpy as np
import pandas as pd
import torch
import torch.nn as nn

from neuralhydrology.datautils.utils import get_frequency_factor, sort_frequencies
from neuralhydrology.modelzoo.basemodel import BaseModel
from neuralhydrology.modelzoo.head import get_head
from neuralhydrology.modelzoo.customlstm import _LSTMCell
from neuralhydrology.utils.config import Config

LOGGER = logging.getLogger(__name__)


class ODELSTM(BaseModel):
    """ODE-LSTM from [#]_.

    An ODE-RNN post-processes the hidden state of a normal LSTM.
    Parts of this code are derived from https://github.com/mlech26l/learning-long-term-irregular-ts.
    
    The forward pass in this model works somewhat differently than the other models, because ODE-LSTM relies on
    irregularly timed samples. To simulate such irregularity, we aggregate parts of the input sequence to random
    frequencies. While doing so, we try to take care that we don't aggregate too coarsely right before the model
    should create a high-frequency prediction.

    Since this aggregation means that parts of the input sequence are at random frequencies, we cannot easily
    return predictions for the full input sequence at each frequency. Instead, we only return sequences of length
    predict_last_n for each frequency (we do not apply the random aggregation to these last time steps).

    The following describes the aggregation strategy implemented in the forward method:

    1. slice one: random-frequency steps (cfg.ode_random_freq_lower_bound <= freq <= lowest-freq) until beginning
                  of the second-lowest frequency input sequence.
    2. slice two: random-frequency steps (lowest-freq <= freq <= self._frequencies[1]) until beginning of
                  next-higher frequency input sequence.
    3. repeat step two until beginning of highest-frequency input sequence.
    4. slice three: random-frequency steps (self._frequencies[-2] <= freq <= highest-freq) until predict_last_n
                    of the lowest frequency.
    5. lowest-frequency steps to generate predict_last_n lowest-frequency predictions.
    6. repeat steps four and five for the next-higher frequency (using the same random-frequency bounds but
       generating predictions for the next-higher frequency).
       
    Parameters
    ----------
    cfg : Config
        The run configuration.

    References
    ----------
    .. [#] Lechner, M.; Hasani, R.: Learning Long-Term Dependencies in Irregularly-Sampled Time Series. arXiv, 2020,
        https://arxiv.org/abs/2006.04418.
    """
    # specify submodules of the model that can later be used for finetuning. Names must match class attributes
    module_parts = ['lstm_cell', 'ode_cell', 'head']

    def __init__(self, cfg: Config):
        super(ODELSTM, self).__init__(cfg=cfg)
        if len(cfg.use_frequencies) < 2:
            raise ValueError('ODELSTM needs at least two frequencies.')
        if isinstance(cfg.dynamic_inputs, dict) or isinstance(cfg.hidden_size, dict):
            raise ValueError('ODELSTM does not support per-frequency input variables or hidden sizes.')

        # Note: be aware that frequency_factors and slice_timesteps have a slightly different meaning here vs. in
        # MTSLSTM. Here, the frequency_factor is relative to the _lowest_ (not the next-lower) frequency.
        # slice_timesteps[freq] is the input step (counting backwards) in the next-*lower* frequency from where on input
        # data at frequency freq is available.
        self._frequency_factors = {}
        self._slice_timesteps = {}
        self._frequencies = sort_frequencies(cfg.use_frequencies)
        self._init_frequency_factors_and_slice_timesteps()

        # start to count the number of inputs
        self.input_size = len(cfg.dynamic_inputs + cfg.static_attributes + cfg.hydroatlas_attributes +
                              cfg.evolving_attributes)

        if cfg.use_basin_id_encoding:
            self.input_size += cfg.number_of_basins
        if cfg.head.lower() == 'umal':
            self.input_size += 1

        self.lstm_cell = _LSTMCell(self.input_size, self.cfg.hidden_size, cfg.initial_forget_bias)
        self.ode_cell = _ODERNNCell(self.cfg.hidden_size,
                                    self.cfg.hidden_size,
                                    num_unfolds=self.cfg.ode_num_unfolds,
                                    method=self.cfg.ode_method)
        self.dropout = nn.Dropout(p=cfg.output_dropout)
        self.head = get_head(cfg=cfg, n_in=self.cfg.hidden_size, n_out=self.output_size)

    def _init_frequency_factors_and_slice_timesteps(self):
        for i, freq in enumerate(self._frequencies):
            frequency_factor = get_frequency_factor(self._frequencies[0], freq)
            if frequency_factor != int(frequency_factor):
                raise ValueError('Frequencies must be multiples of the lowest frequency.')
            self._frequency_factors[freq] = int(frequency_factor)

            if i > 0:
                prev_frequency_factor = get_frequency_factor(self._frequencies[i - 1], freq)
                if self.cfg.predict_last_n[freq] % prev_frequency_factor != 0:
                    raise ValueError(
                        'At all frequencies, predict_last_n must align with the steps of the next-lower frequency.')

                if self.cfg.seq_length[freq] > self.cfg.seq_length[self._frequencies[i - 1]] * prev_frequency_factor:
                    raise ValueError('Higher frequencies must have shorter input sequences than lower frequencies.')

                # we want to pass the state of the day _before_ the next higher frequency starts,
                # because e.g. the mean of a day is stored at the same date at 00:00 in the morning.
                slice_timestep = self.cfg.seq_length[freq] / prev_frequency_factor
                if slice_timestep != int(slice_timestep):
                    raise ValueError('At all frequencies, seq_length must align with the next-lower frequency steps.')
                self._slice_timesteps[freq] = int(slice_timestep)

                # in theory, the following conditions would be possible, but they would make the implementation
                # quite complex and are probably hardly ever useful.
                if self.cfg.predict_last_n[self._frequencies[
                        i - 1]] < self.cfg.predict_last_n[freq] / prev_frequency_factor:
                    raise NotImplementedError(
                        'Lower frequencies cannot have smaller predict_last_n values than higher ones.')

        if any(self.cfg.predict_last_n[f] / self._frequency_factors[f] > self._slice_timesteps[self._frequencies[-1]] /
               self._frequency_factors[self._frequencies[-2]] for f in self._frequencies):
            raise NotImplementedError('predict_last_n cannot be larger than sequence length of highest frequency.')

    def _prepare_inputs(self, data: Dict[str, torch.Tensor],
                        freq: str) -> Union[torch.Tensor, Tuple[torch.Tensor, torch.Tensor]]:
        """Concat all different inputs to the time series input. """
        suffix = f"_{freq}"
        # transpose to [seq_length, batch_size, n_features]
        x_d = data[f'x_d{suffix}'].transpose(0, 1)

        # concat all inputs
        if f'x_s{suffix}' in data and 'x_one_hot' in data:
            x_s = data[f'x_s{suffix}'].unsqueeze(0).repeat(x_d.shape[0], 1, 1)
            x_one_hot = data['x_one_hot'].unsqueeze(0).repeat(x_d.shape[0], 1, 1)
            x_d = torch.cat([x_d, x_s, x_one_hot], dim=-1)
        elif f'x_s{suffix}' in data:
            x_s = data[f'x_s{suffix}'].unsqueeze(0).repeat(x_d.shape[0], 1, 1)
            x_d = torch.cat([x_d, x_s], dim=-1)
        elif 'x_one_hot' in data:
            x_one_hot = data['x_one_hot'].unsqueeze(0).repeat(x_d.shape[0], 1, 1)
            x_d = torch.cat([x_d, x_one_hot], dim=-1)
        else:
            pass

        # add frequency indicator. This will not be used as normal input, but the ODE-RNN uses it to determine the
        # elapsed time since the last sample.
        frequency_factor = 1 / self._frequency_factors[freq]
        frequency_encoding = torch.ones(x_d.shape[0], x_d.shape[1], 1).to(x_d) * frequency_factor
        return torch.cat([x_d, frequency_encoding], dim=-1)

    def _randomize_freq(self, x_d: torch.Tensor, low_frequency: str, high_frequency: str) -> torch.Tensor:
        """Randomize the frequency of the  input sequence. """
        frequency_factor = int(get_frequency_factor(low_frequency, high_frequency))
        possible_aggregate_steps = list(filter(lambda n: frequency_factor % n == 0, range(1, frequency_factor + 1)))

        t = 0
        max_t = x_d.shape[0] / frequency_factor
        x_d_randomized = []
        while t < max_t:
            highfreq_slice = x_d[t * frequency_factor:(t + 1) * frequency_factor]

            # aggregate to a random frequency between low and high
            random_aggregate_steps = np.random.choice(possible_aggregate_steps)
            if highfreq_slice.shape[0] % random_aggregate_steps == 0:
                randfreq_slice = highfreq_slice.view(-1, random_aggregate_steps, highfreq_slice.shape[1],
                                                     highfreq_slice.shape[2]).mean(dim=1)
                # update the frequency indicators.
                randfreq_slice[:, :, -1] = random_aggregate_steps / self._frequency_factors[high_frequency]
            else:
                # do not randomize last slice if it doesn't align with aggregation steps
                randfreq_slice = highfreq_slice
            x_d_randomized.append(randfreq_slice)

            t += 1

        return torch.cat(x_d_randomized, dim=0)

    def forward(self, data: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:
        """Perform a forward pass on the ODE-LSTM model.

        Parameters
        ----------
        data : Dict[str, torch.Tensor]
            Input data for the forward pass. See the documentation overview of all models for details on the dict keys.

        Returns
        -------
        Dict[str, torch.Tensor]
            Model predictions for each target timescale.
        """

        x_d = {freq: self._prepare_inputs(data, freq) for freq in self._frequencies}

        slice_one = self._randomize_freq(x_d[self._frequencies[0]][:-self._slice_timesteps[self._frequencies[1]]],
                                         self.cfg.ode_random_freq_lower_bound, self._frequencies[0])

        batch_size = slice_one.shape[1]
        h_0 = slice_one.data.new(batch_size, self.cfg.hidden_size).zero_()
        c_0 = slice_one.data.new(batch_size, self.cfg.hidden_size).zero_()
        h_n, c_n = self._run_odelstm(slice_one, h_0, c_0)

        prev_freq = self._frequencies[0]
        i = 1
        for freq in self._frequencies[1:-1]:
            to_randomize = x_d[freq][:-self._slice_timesteps[self._frequencies[i + 1]]]

            # slice_timestep of this and the next frequency are identical, so we can move on to the next frequency
            if len(to_randomize) == 0:
                continue
            # random-frequency steps until the beginning of the highest-frequency input sequence.
            slice_two = self._randomize_freq(to_randomize, prev_freq, freq)
            h_n, c_n = self._run_odelstm(slice_two, h_n[:, -1], c_n)
            prev_freq = freq
            i += 1

        pred = {}
        prev_freq_end_step = 0
        for freq in self._frequencies:
            # run random-frequency steps until predict_last_n of this frequency
            end_step = -int(self.cfg.predict_last_n[freq] * self._frequency_factors[self._frequencies[-1]] /
                            self._frequency_factors[freq])
            if end_step == 0:
                continue  # this means the current frequency should not be predicted, so we skip it.

            # check that the slice wouldn't be empty. If it wouldn't, this means that the previous frequency's
            # predict_last_n is not aligned with this frequency's predict_last_n, so we can run random-frequency
            # inputs until the beginning of this frequency's predict_last_n.
            if end_step != prev_freq_end_step and end_step > -self.cfg.seq_length[self._frequencies[-1]]:
                slice_three = self._randomize_freq(x_d[self._frequencies[-1]][prev_freq_end_step:end_step], prev_freq,
                                                   self._frequencies[-1])
                h_n, c_n = self._run_odelstm(slice_three, h_n[:, -1], c_n)
                prev_freq_end_step = end_step

            # run predict_last_n steps at the target frequency
            pred_slice = x_d[freq][-self.cfg.predict_last_n[freq]:]
            h_n_out, _ = self._run_odelstm(pred_slice, h_n[:, -1], c_n)
            pred[f'y_hat_{freq}'] = self.head(self.dropout(h_n_out))['y_hat']

        return pred

    def _run_odelstm(self, input_slice: torch.Tensor, h_0: torch.Tensor,
                     c_0: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        """Ingest `input_slice` into the ODE-LSTM and return hidden states and last cell state. """
        h_x = (h_0, c_0)

        h_n = []
        for x_t in input_slice:
            h_0, c_0 = h_x

            # separate frequency indicator from input variables
            t_input = x_t[:, :-1]
            t_elapsed = x_t[0, -1]

            lstm_out = self.lstm_cell(x_t=t_input, h_0=h_0, c_0=c_0)
            ode_out = self.ode_cell(lstm_out['h_n'], h_0, t_elapsed)

            h_x = (ode_out, lstm_out['c_n'])
            h_n.append(ode_out)

        # stack h to [batch size, sequence length, hidden size]
        return torch.stack(h_n, 1), h_x[1]


class _ODERNNCell(nn.Module):
    """An ODE-RNN cell (Adapted from https://github.com/mlech26l/learning-long-term-irregular-ts) [#]_. 
    
    Parameters
    ----------
    input_size : int
        Input dimension
    hidden_size : int
        Size of the cell's hidden state
    num_unfolds : int
        Number of steps into which each timestep will be broken down to solve the ODE.
    method : {'euler', 'heun', 'rk4'}
        Method to use for ODE solving (Euler's method, Heun's method, or Runge-Kutta 4)
    
    References
    ----------
    .. [#] Lechner, M.; Hasani, R.: Learning Long-Term Dependencies in Irregularly-Sampled Time Series. arXiv, 2020,
        https://arxiv.org/abs/2006.04418.
    """

    def __init__(self, input_size: int, hidden_size: int, num_unfolds: int, method: str):
        super(_ODERNNCell, self).__init__()
        self.method = {
            'euler': self._euler,
            'heun': self._heun,
            'rk4': self._rk4,
        }[method]
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.num_unfolds = num_unfolds

        self.w_ih = nn.Parameter(torch.FloatTensor(hidden_size, input_size))
        self.w_hh = nn.Parameter(torch.FloatTensor(hidden_size, hidden_size))
        self.bias = nn.Parameter(torch.FloatTensor(hidden_size))
        self.scale = nn.Parameter(torch.FloatTensor(hidden_size))

        self._reset_parameters()

    def _reset_parameters(self):
        """Reset the paramters of the ODERNNCell. """
        nn.init.orthogonal_(self.w_hh)
        nn.init.xavier_uniform_(self.w_ih)
        nn.init.zeros_(self.bias)
        nn.init.constant_(self.scale, 1.0)

    def forward(self, new_hidden_state: torch.Tensor, old_hidden_state: torch.Tensor, elapsed: float) -> torch.Tensor:
        """Perform a forward pass on the ODERNNCell.
        
        Parameters
        ----------
        new_hidden_state : torch.Tensor
            The current hidden state to be updated by the ODERNNCell.
        old_hidden_state : torch.Tensor
            The previous hidden state.
        elapsed : float
            Time elapsed between new and old hidden state.

        Returns
        -------
        torch.Tensor
            Predicted new hidden state
        """
        delta_t = elapsed / self.num_unfolds

        hidden_state = old_hidden_state
        for i in range(self.num_unfolds):
            hidden_state = self.method(new_hidden_state, hidden_state, delta_t)
        return hidden_state

    def _dfdt(self, inputs: torch.Tensor, hidden_state: torch.Tensor) -> torch.Tensor:
        h_in = torch.matmul(inputs, self.w_ih)
        h_rec = torch.matmul(hidden_state, self.w_hh)
        dh_in = self.scale * torch.tanh(h_in + h_rec + self.bias)
        dh = dh_in - hidden_state
        return dh

    def _euler(self, inputs: torch.Tensor, hidden_state: torch.Tensor, delta_t: float) -> torch.Tensor:
        dy = self._dfdt(inputs, hidden_state)
        return hidden_state + delta_t * dy

    def _heun(self, inputs: torch.Tensor, hidden_state: torch.Tensor, delta_t: float) -> torch.Tensor:
        k1 = self._dfdt(inputs, hidden_state)
        k2 = self._dfdt(inputs, hidden_state + delta_t * k1)
        return hidden_state + delta_t * 0.5 * (k1 + k2)

    def _rk4(self, inputs: torch.Tensor, hidden_state: torch.Tensor, delta_t: float) -> torch.Tensor:
        k1 = self._dfdt(inputs, hidden_state)
        k2 = self._dfdt(inputs, hidden_state + k1 * delta_t * 0.5)
        k3 = self._dfdt(inputs, hidden_state + k2 * delta_t * 0.5)
        k4 = self._dfdt(inputs, hidden_state + k3 * delta_t)

        return hidden_state + delta_t * (k1 + 2 * k2 + 2 * k3 + k4) / 6.0
